# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
from django.conf import settings


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('hub', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='task',
            name='owner',
            field=models.ForeignKey(to=settings.AUTH_USER_MODEL),
        ),
        migrations.AddField(
            model_name='task',
            name='parent',
            field=models.ForeignKey(blank=True, to='hub.Task', help_text='Parent task.', null=True),
        ),
        migrations.AddField(
            model_name='task',
            name='resubmitted_by',
            field=models.ForeignKey(related_name='resubmitted_by1', blank=True, to=settings.AUTH_USER_MODEL, null=True),
        ),
        migrations.AddField(
            model_name='task',
            name='resubmitted_from',
            field=models.ForeignKey(related_name='resubmitted_from1', blank=True, to='hub.Task', null=True),
        ),
        migrations.AddField(
            model_name='task',
            name='worker',
            field=models.ForeignKey(blank=True, to='hub.Worker', help_text='A worker which has this task assigned.', null=True),
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
import kobo.django.fields


class Migration(migrations.Migration):

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Arch',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('name', models.CharField(help_text='i386, ia64, ...', unique=True, max_length=16)),
                ('pretty_name', models.CharField(help_text='i386, Itanium, ...', unique=True, max_length=64)),
            ],
            options={
                'ordering': ('name',),
                'verbose_name_plural': 'arches',
            },
        ),
        migrations.CreateModel(
            name='Channel',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('name', models.CharField(help_text='Channel name', max_length=128)),
            ],
        ),
        migrations.CreateModel(
            name='Task',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('archive', models.BooleanField(default=False, help_text="When a task is archived, it disappears from admin interface and cannot be accessed by taskd.<br />Make sure that archived tasks are finished and you won't need them anymore.")),
                ('state', models.PositiveIntegerField(default=0, help_text='Current task state.', choices=[(0, b'FREE'), (1, b'ASSIGNED'), (2, b'OPEN'), (3, b'CLOSED'), (4, b'CANCELED'), (5, b'FAILED'), (6, b'INTERRUPTED'), (7, b'TIMEOUT'), (8, b'CREATED')])),
                ('label', models.CharField(help_text='Label, description or any reason for this task.', max_length=255, blank=True)),
                ('exclusive', models.BooleanField(default=False, help_text='Exclusive tasks have highest priority. They are used e.g. when shutting down a worker.')),
                ('method', models.CharField(help_text='Method name represents appropriate task handler.', max_length=255)),
                ('args', kobo.django.fields.JSONField(default={}, help_text='Method arguments. JSON serialized dictionary.', blank=True)),
                ('result', models.TextField(help_text='Task result. Do not store a lot of data here (use HubProxy.upload_task_log instead).', blank=True)),
                ('comment', models.TextField(null=True, blank=True)),
                ('timeout', models.PositiveIntegerField(help_text='Task timeout. Leave blank for no timeout.', null=True, blank=True)),
                ('waiting', models.BooleanField(default=False, help_text='Task is waiting until some subtasks finish.')),
                ('awaited', models.BooleanField(default=False, help_text='Task is awaited by another task.')),
                ('dt_created', models.DateTimeField(auto_now_add=True)),
                ('dt_started', models.DateTimeField(null=True, blank=True)),
                ('dt_finished', models.DateTimeField(null=True, blank=True)),
                ('priority', models.PositiveIntegerField(default=10, help_text='Priority.')),
                ('weight', models.PositiveIntegerField(default=1, help_text='Weight determines how many resources is used when processing the task.')),
                ('subtask_count', models.PositiveIntegerField(default=0, help_text='Subtask count.<br />This is a generated field.')),
                ('arch', models.ForeignKey(to='hub.Arch')),
                ('channel', models.ForeignKey(to='hub.Channel')),
            ],
            options={
                'ordering': ('-id',),
            },
        ),
        migrations.CreateModel(
            name='Worker',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('worker_key', models.CharField(help_text='Worker authentication key.<br />Leave blank to generate new key.', unique=True, max_length=255, blank=True)),
                ('name', models.CharField(help_text='Worker hostname.', unique=True, max_length=128)),
                ('enabled', models.BooleanField(default=True, help_text='Enabled workers are allowed to process tasks.')),
                ('max_load', models.PositiveIntegerField(default=1, help_text='Maximum allowed load (sum of task weights).', blank=True)),
                ('max_tasks', models.PositiveIntegerField(default=0, help_text='Maximum assigned tasks. (0 = no limit)', blank=True)),
                ('ready', models.BooleanField(default=True, help_text='Is the worker ready to take new tasks?<br />This is a generated field.')),
                ('task_count', models.PositiveIntegerField(default=0, help_text='Count of processed tasks.<br />This is a generated field.', blank=True)),
                ('current_load', models.PositiveIntegerField(default=0, help_text='Sum of task weights.<br />This is a generated field.', blank=True)),
                ('arches', models.ManyToManyField(help_text='Supported architectures', to='hub.Arch')),
                ('channels', models.ManyToManyField(to='hub.Channel')),
            ],
        ),
    ]



#!/usr/bin/python
# -*- coding: utf-8 -*-


import unittest
import run_tests # set sys.path

import tempfile
import os

from kobo.decorators import *


class TestDecoratorsModule(unittest.TestCase):
    def setUp(self):
        self.tmp_file = tempfile.mktemp()

    def tearDown(self):
        os.remove(self.tmp_file)

    def test_log_traceback(self):
        @log_traceback(self.tmp_file)
        def foo_function():
            raise IOError("Some error")

        try:
            foo_function()
        except IOError:
            pass

        tb = open(self.tmp_file).read()
        self.assertTrue(tb.startswith("--- TRACEBACK BEGIN:"))


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-


from kobo.django.menu import MenuItem


menu = (
    MenuItem("Arches", "arch/list"),
    MenuItem("Channels", "channel/list"),
    MenuItem("Users", "user/list"),
    MenuItem("Workers", "worker/list"),
)



# -*- coding: utf-8 -*-


import os
import random

from kobo.client.constants import TASK_STATES
from kobo.hub.models import Task
from kobo.hub.decorators import validate_worker
from kobo.xmlrpc import decode_xmlrpc_chunk


__all__ = (
    "assign_task",
    "open_task",
    "close_task",
    "cancel_task",
    "fail_task",
    "interrupt_tasks",
    "timeout_tasks",

    "get_tasks_to_assign",
    "get_awaited_tasks",
    "get_worker_info",
    "get_worker_id",
    "get_worker_tasks",
    "get_task",
    "get_task_no_verify",

    "set_task_weight",
    "update_worker",
    "create_subtask",
    "wait",
    "check_wait",
    "upload_task_log",
)


@validate_worker
def get_worker_info(request):
    """
    Get information about a worker.

    @rtype: dict
    """
    return request.worker.export()


@validate_worker
def get_worker_id(request):
    """Get worker ID of a worker.
        @return: int
    """
    return request.worker.id


@validate_worker
def get_worker_tasks(request):
    """
    Get list of tasks running on a worker.

    @rtype: list
    """
    task_list = []
    for task in request.worker.running_tasks().order_by("-exclusive", "-awaited", "id"):
        task_info = task.export()

        # set wakeup alert
        if task.waiting:
            finished, unfinished = task.check_wait()
            if len(finished) > 0:
                task_info["alert"] = True

        task_list.append(task_info)
    return task_list


@validate_worker
def get_task(request, task_id):
    """
    Get information about a task.

    @param task_id: a task ID
    @type  task_id: int
    @rtype: dict
    """
    task = Task.objects.get_and_verify(task_id=task_id, worker=request.worker)
    return task.export()


@validate_worker
def get_task_no_verify(request, task_id):
    """
    Get information about a task, do not verify whether is assigned to a worker.

    @param task_id: a task ID
    @type  task_id: int
    @rtype: dict
    """
    task = Task.objects.get(id=task_id)
    return task.export()


@validate_worker
def interrupt_tasks(request, task_list):
    result = True
    for task_id in task_list:
        task = Task.objects.get_and_verify(task_id=task_id, worker=request.worker)
        if task:
            try:
                task.interrupt_task(recursive=True)
            except:
                raise
                result = False
    return result


@validate_worker
def timeout_tasks(request, task_list):
    result = True
    for task_id in task_list:
        task = Task.objects.get_and_verify(task_id=task_id, worker=request.worker)
        if task:
            try:
                task.timeout_task(recursive=True)
            except:
                raise
                result = False
    return result


@validate_worker
def assign_task(request, task_id):
    task = Task.objects.get(id=task_id)
    return task.assign_task(request.worker.id)


@validate_worker
def open_task(request, task_id):
    task = Task.objects.get(id=task_id)
    return task.open_task(request.worker.id)


@validate_worker
def close_task(request, task_id, task_result):
    task = Task.objects.get_and_verify(task_id=task_id, worker=request.worker)
    return task.close_task(task_result=task_result)


@validate_worker
def cancel_task(request, task_id):
    task = Task.objects.get_and_verify(task_id=task_id, worker=request.worker)
    return task.cancel_task()


@validate_worker
def fail_task(request, task_id, task_result):
    task = Task.objects.get_and_verify(task_id=task_id, worker=request.worker)
    return task.fail_task(task_result=task_result)


@validate_worker
def set_task_weight(request, task_id, weight):
    task = Task.objects.get_and_verify(task_id=task_id, worker=request.worker)
    task.setWeight(weight)
    return task.weight


@validate_worker
def update_worker(request, enabled, ready, task_count):
    return request.worker.update_worker(enabled, ready, task_count)


@validate_worker
def get_tasks_to_assign(request):
    task_list = []
    max_tasks = max(request.worker.max_tasks, 10) # return info about at least 10 tasks

    # Limit each query by max_tasks.
    # Worker sometimes doesn't succeed in taking all tasks from task_list,
    # but it can take another tasks next time. This can be good for load balancing
    # (not all tasks are taken by one worker).
    # If task_list is longer than max_tasks, return it not to perform another queries.

    # exclusive tasks
    for task in request.worker.assigned_tasks().filter(exclusive=True).order_by("-priority", "id")[:max_tasks]:
        task_info = task.export(flat=False)
        task_list.append(task_info)

    if len(task_list) >= max_tasks:
        return task_list

    # awaited tasks
    for task in Task.objects.free().filter(awaited=True).order_by("-priority", "id")[:max_tasks]:
        task_info = task.export(flat=False)
        task_list.append(task_info)

    if len(task_list) >= max_tasks:
        return task_list

    # tasks assigned to this worker
    for task in request.worker.assigned_tasks().filter(exclusive=False).order_by("-priority", "id")[:max_tasks]:
        task_info = task.export(flat=False)
        task_list.append(task_info)

    if len(task_list) >= max_tasks:
        return task_list

    # free tasks for each channel relevant to the worker
    tasks = []
    for channel in request.worker.channels.all():
        for task in Task.objects.free().filter(awaited=False, channel=channel, priority__gte=request.worker.min_priority).order_by("-priority", "id")[:max_tasks]:
            task_info = task.export(flat=False)
            tasks.append(task_info)

    # Shuffle the list to prevent task starvation in some channels.
    # It could also help to lower task assignment conflicts.
    # After that, sort it by priority again.
    random.shuffle(tasks)
    tasks.sort(lambda x, y: cmp(x["priority"], y["priority"]), reverse=True)
    task_list.extend(tasks[:max_tasks])

    return task_list


@validate_worker
def get_awaited_tasks(request, awaited_task_list):
    task_list = []
    for task in Task.objects.filter(awaited=True, parent__in=[ i["id"] for i in awaited_task_list ]):#.order_by("-exclusive", "-awaited", "id")[:50]:
        task_info = task.export()
        task_list.append(task_info)
    return task_list


@validate_worker
def create_subtask(request, label, method, args, parent_id):
    parent_task = Task.objects.get_and_verify(task_id=parent_id, worker=request.worker)
#    def create_task(cls, owner_name, label, method, args=None, parent_id=None, worker_name=None, arch_name="noarch", channel_name="default", priority=10, weight=1, exclusive=False):
#    subtask_id = self.__hub.worker.createSubtask(label, method, args, self.__task_id)
    return Task.create_task(parent_task.owner.username, label, method, args=args, parent_id=parent_id, arch_name=parent_task.arch.name, channel_name=parent_task.channel.name)# priority=priority, weight=weight)


@validate_worker
def wait(request, task_id, child_list=None):
    task = Task.objects.get(id=task_id)
    task.wait(child_list)
    return True


@validate_worker
def check_wait(request, task_id, child_list=None):
    task = Task.objects.get(id=task_id)
    return task.check_wait(child_list)


@validate_worker
def upload_task_log(request, task_id, relative_path, mode, chunk_start, chunk_len, chunk_checksum, encoded_chunk):
    """
    Upload a task log.

    @param task_id: task ID
    @type  task_id: int
    @param relative_path: relative path (under task_dir) to the log file
    @type  relative_path: str
    @param mode: file perms (example: 0644)
    @type  mode: int
    @param chunk_start: chunk start position in the file (-1 for append)
    @type  chunk_start: str
    @param chunk_len: chunk length
    @type  chunk_len: str
    @param chunk_checksum: sha256 checksum (lower case)
    @type  chunk_checksum: str
    @param encoded_chunk: base64 encoded chunk
    @type  encoded_chunk: str
    @rtype: bool
    """

    relative_path = os.path.normpath(relative_path)
    if relative_path.startswith(".."):
        raise ValueError("Invalid upload path: %s" % relative_path)

    task = Task.objects.get(id=task_id)
    full_path = os.path.join(task.task_dir(), relative_path)
    if task.state != TASK_STATES["OPEN"]:
        raise ValueError("Can't upload file for a task which is not OPEN: %s" % task_id)

    try:
        decode_xmlrpc_chunk(chunk_start, chunk_len, chunk_checksum, encoded_chunk, write_to=full_path, mode=mode)
    except ValueError:
        return False

    return True



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import unittest
from StringIO import StringIO

from kobo.hub.models import _tail as tail


SAMPLE_LINES = [
    'this is a sample',
    'string;     each',
    'line    contains',
    '16 chars exclud-',
    'ing newline, and',
    'six lines  total'
]
SAMPLE_STRING = '\n'.join(SAMPLE_LINES)


class TestTail(unittest.TestCase):

    def test_tail_empty(self):
        """tail of empty object returns empty"""
        (actual, offset) = tail(StringIO(), 1024, 1024)

        self.assertEqual(actual, '')
        self.assertEqual(offset, 0)

    def test_tail_noop(self):
        """tail returns all content if it fits in requested size"""
        (actual, offset) = tail(StringIO(SAMPLE_STRING), 1024, 1024)

        self.assertEqual(actual, SAMPLE_STRING)
        self.assertEqual(offset, len(SAMPLE_STRING))

    def test_tail_limit(self):
        """tail returns trailing lines up to given limit"""
        expected = '\n'.join([
            'ing newline, and',
            'six lines  total',
        ])
        (actual, offset) = tail(StringIO(SAMPLE_STRING), 40, 1024)

        self.assertEqual(actual, expected)
        self.assertEqual(offset, len(SAMPLE_STRING))

    def test_tail_line_break(self):
        """tail breaks in middle of line if lines are longer than max length"""
        expected = '\n'.join([
            # this line is partially returned
            'xclud-',
            'ing newline, and',
            'six lines  total',
        ])
        (actual, offset) = tail(StringIO(SAMPLE_STRING), 40, 10)

        self.assertEqual(actual, expected)
        self.assertEqual(offset, len(SAMPLE_STRING))


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-


import sys

from kobo.client import ClientCommand


class Cancel_Tasks(ClientCommand):
    """cancel free, assigned or open tasks"""
    enabled = True


    def options(self):
        self.parser.usage = "%%prog %s task_id [task_id...]" % self.normalized_name


    def run(self, *args, **kwargs):
        if len(args) == 0:
            self.parser.error("At least one task id must be specified.")

        username = kwargs.pop("username", None)
        password = kwargs.pop("password", None)
        tasks = args

        self.set_hub(username, password)

        failed = False
        for task_id in tasks:
            try:
                result = self.hub.client.cancel_task(task_id)
                if result and isinstance(result, basestring):
                    print result
            except Exception, ex:
                failed = True
                print ex

        if failed:
            sys.exit(1)



# -*- coding: utf-8 -*-


import os

import kobo.cli
import kobo.admin


class Start_Worker(kobo.cli.Command):
    """create a worker directory structure in the current directory"""
    enabled = True

    def options(self):
        self.parser.usage = "%%prog %s [options] <project_name>" % self.normalized_name

    def run(self, *args, **kwargs):
        if len(args) < 1:
            self.parser.error("Please specify a name of the project.")

        name = args[0]
        directory = os.getcwd()

        try:
            kobo.admin.copy_helper(name, directory, "worker")
        except kobo.admin.TemplateError, ex:
            self.parser.error(ex)



#!/usr/bin/python
# -*- coding: utf-8 -*-

import sys
import os

PROJECT_DIR = os.path.abspath(os.path.join(os.path.dirname(__file__), ".."))
os.environ["PYTHONPATH"] = PROJECT_DIR
os.environ['DJANGO_SETTINGS_MODULE'] = 'fields_test.settings'
sys.path.insert(0, PROJECT_DIR)

from django.core.management import call_command
import django

# Django >= 1.7 must call this method to initialize app registry,
# while older Django do not have this method
if 'setup' in dir(django):
    django.setup()

call_command('test', 'fields_test')
#call_command('syncdb')



# -*- coding: utf-8 -*-


import django.forms as forms
from django.db.models import Q
from kobo.hub.models import Task


class TaskSearchForm(forms.Form):
    search  = forms.CharField(required=False)
    my  = forms.BooleanField(required=False)

    def __init__(self, *args, **kwargs):
        self.state = kwargs.pop('state', None)
        self.order_by = kwargs.pop('order_by', ['-id'])
        return super(TaskSearchForm, self).__init__(*args, **kwargs)

    def get_query(self, request):
        self.is_valid()
        search = self.cleaned_data["search"]
        my = self.cleaned_data["my"]

        query = Q()

        if search:
            query |= Q(method__icontains=search)
            query |= Q(owner__username__icontains=search)
            query |= Q(label__icontains=search)

        if my and request.user.is_authenticated():
            query &= Q(owner=request.user)

        if self.state is not None:
            query &= Q(state__in=self.state)
        #if self.kwargs:
        #    query &= Q(self.kwargs)
        return Task.objects.filter(parent__isnull=True).filter(query).order_by(*self.order_by).defer("result", "args").select_related("owner", "worker")



# -*- coding: utf-8 -*-


"""
Various useful shortcuts.
"""


import os
import sys
import subprocess
import random
import re
import hashlib
import threading


__all__ = (
    "force_list",
    "force_tuple",
    "allof",
    "anyof",
    "noneof",
    "oneof",
    "is_empty",
    "iter_chunks",
    "random_string",
    "hex_string",
    "touch",
    "read_from_file",
    "save",
    "save_to_file",
    "find_symlinks_to",
    "run",
    "compute_file_checksums",
    "parse_checksum_line",
    "read_checksum_file",
    "split_path",
    "relative_path",
    "makedirs",
)


def force_list(value):
    """Convert a value to list.

    @rtype: list
    """
    if type(value) is list:
        return value
    if type(value) in (tuple, set):
        return list(value)
    return [value]


def force_tuple(value):
    """Convert a value to tuple.

    @rtype: tuple
    """
    if type(value) is tuple:
        return value
    if type(value) in (list, set):
        return tuple(value)
    return (value, )


def allof(*args, **kwargs):
    """Test if all values are True.

    @rtype: bool
    """
    for i in list(args) + kwargs.values():
        if not i:
            return False
    return True


def anyof(*args, **kwargs):
    """Test if at least one of the values is True.

    @rtype: bool
    """
    for i in list(args) + kwargs.values():
        if i:
            return True
    return False


def noneof(*args, **kwargs):
    """Test if all values are False.

    @rtype: bool
    """
    for i in list(args) + kwargs.values():
        if i:
            return False
    return True


def oneof(*args, **kwargs):
    """Test if just one of the values is True.

    @rtype: bool
    """
    found = False
    for i in list(args) + kwargs.values():
        if i:
            if found:
                return False
            found = True
    return found


def is_empty(value):
    """Test if value is None, empty string, list, tuple or dict.

    @rtype: bool
    """
    if value is None:
        return True
    if type(value) in (list, tuple, dict):
        return len(value) == 0
    if not value:
        return True
    return False


def iter_chunks(input_list, chunk_size):
    """Iterate through input_list and yield chunk_size-d chunks."""

    # TODO: any idea how to detect files better? This works for StringIO at least...
    is_file = type(input_list) is file
    is_file |= hasattr(input_list, "read") and hasattr(input_list, "close") and hasattr(input_list, "seek")
    if is_file:
        while 1:
            chunk = input_list.read(chunk_size)
            if not chunk:
                break
            yield chunk
        return

    # TODO: any idea how to detect generators better?
    if hasattr(input_list, "__iter__") and hasattr(input_list, "next"):
        chunk = []
        for i in input_list:
            chunk.append(i)
            if len(chunk) == chunk_size:
                yield chunk
                chunk = []
        if chunk:
            yield chunk
        return

    can_slice = hasattr(input_list, "__getslice__")
    for i in xrange(0, len(input_list), chunk_size):
        if can_slice:
            # regular list
            yield input_list[i:i + chunk_size]
        else:
            # xrange, etc.
            chunk = []
            for j in xrange(i, i + chunk_size):
                if j < len(input_list):
                    chunk.append(input_list[j])
            yield chunk


def random_string(length=32, alphabet=None):
    """Return random string of given lenght which consists of characters from the alphabet.

    @rtype: str
    """
    alphabet = alphabet or "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789"
    return "".join([ random.choice(alphabet) for i in xrange(length) ])


def hex_string(string):
    """Convert a string to a string of hex digits.

    @rtype: str
    """
    return "".join(( "%02x" % ord(i) for i in string ))


def touch(filename, mode=0644):
    """Touch a file."""
    save_to_file(filename, "", append=True, mode=mode)


def read_from_file(filename, lines=None, re_filter=None):
    """Read a text file."""
    result = []
    fo = open(filename, "rt")

    if re_filter is not None:
        re_filter_compiled = re.compile(re_filter)

    for i, line in enumerate(fo):
        if lines is not None and i+1 not in lines:
            continue
        line = line.rstrip("\r\n")
        if re_filter is not None and not re_filter_compiled.match(line):
            continue
        result.append(line)

    fo.close()
    return result


def save_to_file(filename, text, append=False, mode=0644):
    """Save text to a file."""
    if append and os.path.exists(filename):
        fd = os.open(filename, os.O_RDWR | os.O_APPEND, mode)
    else:
        fd = os.open(filename, os.O_WRONLY | os.O_CREAT | os.O_TRUNC, mode)
    os.write(fd, text)
    os.close(fd)


def save(*args, **kwargs):
    print >> sys.stderr, "DeprecationWarning: kobo.shortcuts.save() is deprecated, use kobo.shortcuts.save_to_file() instead."
    return save_to_file(*args, **kwargs)


def find_symlinks_to(target, directory):
    """Find symlinks which point to a target.

    @param target: the symlink target we're looking for
    @type target: str
    @param directory: directory with symlinks
    @type directory: str
    @return: list of symlinks to the target
    @rtype: list
    """

    target = os.path.abspath(target)
    directory = os.path.abspath(directory)
    result = []
    for fn in os.listdir(directory):
        path = os.path.abspath(os.path.join(directory, fn))

        if not os.path.islink(path):
            continue

        if os.path.abspath(os.path.join(directory, os.path.normpath(os.readlink(path)))) == os.path.abspath(target):
            result.append(path)

    return result


def run(cmd, show_cmd=False, stdout=False, logfile=None, can_fail=False, workdir=None, stdin_data=None, return_stdout=True, buffer_size=4096, **kwargs):
    """Run a command in shell.

    @param show_cmd: show command in stdout/log
    @type show_cmd: bool
    @param stdout: print output to stdout
    @type stdout: bool
    @param logfile: save output to logfile
    @type logfile: str
    @param can_fail: when set, retcode is returned instead of raising RuntimeError
    @type can_fail: bool
    @param workdir: change current directory to workdir before starting a command
    @type workdir: str
    @param stdin_data: stdin data passed to a command
    @type stdin_data: str
    @param buffer_size: size of buffer for reading from proc's stdout, use -1 for line-buffering
    @type buffer_size: int
    @return_stdout: return command stdout as a function result (turn off when working with large data, None is returned instead of stdout)
    @return_stdout: bool
    @return: (command return code, merged stdout+stderr)
    @rtype: (int, str) or (int, None)
    """
    if type(cmd) in (list, tuple):
        import pipes
        cmd = " ".join((pipes.quote(i) for i in cmd))

    log = None
    if logfile:
        logfile = os.path.join(workdir or "", logfile)
        # What happens to log file if it exists already? If show_cmd is True,
        # it will be overwritten. Otherwise the command output will just be
        # appended to the existing file.
        mode = 'a' if not show_cmd and os.path.exists(logfile) else 'w'
        log = open(logfile, mode)

    try:

        if show_cmd:
            command = "COMMAND: %s\n%s\n" % (cmd, "-" * (len(cmd) + 9))
            if stdout:
                print command,
            if logfile:
                log.write(command)

        stdin = None
        if stdin_data is not None:
            stdin = subprocess.PIPE

        proc = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE,
                                stderr=subprocess.STDOUT, stdin=stdin,
                                cwd=workdir, **kwargs)

        if stdin_data is not None:
            class StdinThread(threading.Thread):
                def run(self):
                    proc.stdin.write(stdin_data)
                    proc.stdin.close()
            stdin_thread = StdinThread()
            stdin_thread.daemon = True
            stdin_thread.start()

        output = ""
        while True:
            if buffer_size == -1:
                lines = proc.stdout.readline()
            else:
                try:
                    lines = proc.stdout.read(buffer_size)
                except (IOError, OSError), ex:
                    import errno
                    if ex.errno == errno.EINTR:
                        continue
                    else:
                        raise

            if lines == "":
                break
            if stdout:
                sys.stdout.write(lines)
            if logfile:
                log.write(lines)
            if return_stdout:
                output += lines
        proc.wait()

    finally:
        if logfile:
            log.close()

    if stdin_data is not None:
        stdin_thread.join()

    err_msg = "ERROR running command: %s" % cmd
    if logfile:
        err_msg += "\nFor more details see %s" % logfile

    if proc.returncode != 0 and not can_fail:
        exc = RuntimeError(err_msg)
        exc.output = output
        raise exc

    if proc.returncode != 0 and show_cmd:
        print >> sys.stderr, err_msg

    if not return_stdout:
        output = None

    return (proc.returncode, output)


def compute_file_checksums(filename, checksum_types):
    """
    Compute a file checksum of given type.
    Checksum must be supported by hashlib.

    @param filename: path to a file
    @type filename: str
    @param checksum_type: checksum types, supported by hashlib
    @type checksum_type: str or list
    @return: {checksum_type: digest in lowercase hex}
    @rtype: dict
    """

    checksum_types = set(force_tuple(checksum_types))
    checksums = {}

    for checksum_type in checksum_types:
        try:
            checksums[checksum_type] = getattr(hashlib, checksum_type)()
        except AttributeError:
            raise ValueError("Checksum is not supported in hashlib: %s" % checksum_type)

    fo = open(filename, "r")
    while True:
        chunk = fo.read(1024**2)
        if not chunk:
            break
        for checksum_type in checksum_types:
            checksums[checksum_type].update(chunk)
    fo.close()

    result = {}
    for checksum_type, checksum in checksums.iteritems():
        result[checksum_type] = checksum.hexdigest().lower()
    return result


CHECKSUM_LINE_RE = re.compile(r"^(?P<escaped>\\)?(?P<checksum>\S+) [ \*](?P<path>.*)$")
def parse_checksum_line(line):
    """Parse a line of md5sum, sha256sum, ... file.

    @param line: line of a *sum file
    @type line: str
    @return: (checksum, path)
    @rtype: (str, str)
    """

    if not line.strip():
        return None
    if line.strip().startswith("#"):
        return None

    match = CHECKSUM_LINE_RE.match(line)
    data = match.groupdict()
    if data["escaped"]:
        data["path"] = data["path"].decode('string-escape')
    return (data["checksum"], data["path"])


def read_checksum_file(file_name):
    """Read checksums from a file.

    @param file_name: checksum file
    @type file_name: str
    @return: [(checksum, path)]
    @rtype: [(str, str)]
    """

    result = []
    fo = open(file_name, "rb")

    for line in fo:
        checksum_tuple = parse_checksum_line(line)
        if checksum_tuple is not None:
            result.append(checksum_tuple)

    fo.close()
    return result


def split_path(path):
    """Split path to a list.

    @param path: a path
    @type path: str
    @return: list with path elements
    @rtype: list
    """

    head, tail = os.path.split(os.path.normpath(path))
    if not head:
        return [tail]
    if not tail:
        if head == os.sep:
            return [head]
        return split_path(head[:-len(os.sep)])
    return split_path(head) + [tail]


def relative_path(src_path, dst_path=None):
    """Determine relative path between two paths.
    This function is primarily meant for tweaking paths for symlinks.
    Paths ending with '/' are considered as directories.

    @param src_path:
    @type src_path:
    @param: dst_path:
    @type dst_path:
    @returns: relative path from src to dst
    @rtype: str
    """

    if dst_path is None:
        dst_path = os.path.curpath(os.curdir) + "/"
    src_path, src_file = os.path.split(src_path)
    dst_path, dst_file = os.path.split(dst_path)

    src_path = split_path(src_path)
    dst_path = split_path(dst_path)

    if not src_file and dst_file:
        raise RuntimeError("")

    while src_path and dst_path and src_path[0] == dst_path[0]:
        src_path.pop(0)
        dst_path.pop(0)

    while dst_path:
        src_path.insert(0, "..")
        dst_path.pop(0)

    src_path.append(src_file)
    dst_path.append(dst_file)
    return os.path.join(*src_path)


def makedirs(path, mode=0777):
    """
    Wrapper to os.makedirs which does not
    throw an exception on existing directory.
    """
    try:
        os.makedirs(path, mode)
    except OSError, ex:
        if ex.errno != 17:
            raise
        if not os.path.isdir(path):
            raise



# -*- coding: utf-8 -*-

# Copyright (c) 2009 Red Hat, Inc.
#
# Copyright (c) Django Software Foundation and individual contributors.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright notice,
#        this list of conditions and the following disclaimer.
#
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#
#     3. Neither the name of Django nor the names of its contributors may be used
#        to endorse or promote products derived from this software without
#        specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
# ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


import os
import sys
import traceback
import re


__all__ = (
    "Traceback",
    "get_traceback",
    "set_except_hook",
)


def get_traceback():
    """Return a traceback string."""
    exc_info = sys.exc_info()
    if exc_info[0] is None:
        return ""
    return "".join(traceback.format_exception(*exc_info)).replace(r"\n", "\n")


def get_exception():
    """Return an exception string."""
    result = sys.exc_info()[1]
    if result is None:
        return ""
    return str(result)


class Traceback(object):
    """Enhanced traceback with detailed output."""

    def __init__(self, exc_info=None, show_traceback=True, show_code=True, show_locals=True, show_environ=False, show_modules=False):
        self.exc_info = exc_info or sys.exc_info()
        self.show_traceback = show_traceback
        self.show_code = show_code
        self.show_locals = show_locals
        self.show_environ = show_environ
        self.show_modules = show_modules

    def _to_str(self, value, format=None):
        """Convert value to string.

        We must absolutely avoid propagating exceptions, and str(value)
        COULD cause any exception, so we MUST catch any...
        """

        format = format or "%s"
        try:
            result = format % value
        except:
            result = "<ERROR WHILE CONVERTING VALUE TO STRING>"
        return result

    def get_traceback(self):
        """Return a traceback string."""
        if self.exc_info[0] is None:
            return ""
        result = []

        if self.show_traceback:
            c_pattern = re.compile('\n')
            for i in traceback.format_exception(*self.exc_info):
                for line in c_pattern.split(i):
                    line and result.append(line)

        if self.show_environ:
            result.append("<ENVIRON>")
            for key, value in sorted(os.environ.iteritems()):
                result.append("%s = %s" % (self._to_str(key, "%20s"), self._to_str(value)))
            result.append("</ENVIRON>")

        if self.show_environ:
            result.append("<GLOBALS>")
            for key, value in sorted(os.environ.iteritems()):
                result.append("%s = %s" % (self._to_str(key, "%20s"), self._to_str(value)))
            result.append("</GLOBALS>")

        if self.show_modules:
            result.append("<MODULES>")
            for key, value in sorted(sys.modules.iteritems()):
                result.append("%s = %s" % (self._to_str(key, "%20"), self._to_str(value)))
            result.append("</MODULES>")

        if self.show_code or self.show_locals:
            for frame in reversed(self.get_frames()):
                if frame["filename"] == "?" or frame["lineno"] == "?":
                    continue

                result.append("Frame %s in %s at line %s" % (
                    self._to_str(frame["function"]),
                    self._to_str(frame["filename"]),
                    self._to_str(frame["lineno"]))
                )

                if self.show_code:
                    result.append("<CODE>")
                    lineno = frame["pre_context_lineno"]
                    for line in frame["pre_context"]:
                        result.append("    %s %s" % (self._to_str(lineno, "%4d"), self._to_str(line)))
                        lineno += 1

                    result.append("--> %s %s" % (self._to_str(lineno, "%4d"), self._to_str(frame["context_line"])))
                    lineno += 1

                    for line in frame["post_context"]:
                        result.append("    %s %s" % (self._to_str(lineno, "%4d"), self._to_str(line)))
                        lineno += 1
                    result.append("</CODE>")

                if self.show_locals:
                    result.append("<LOCALS>")
                    for key, value in sorted(frame["vars"]):
                        result.append("%20s = %s" % (self._to_str(key), self._to_str(value, "%.200r")))
                        if key == "self":
                            try:
                                variables = sorted(dir(value))
                            except Exception, ex:
                                # this exception may be thrown on xmlrpc proxy object
                                # e.g. <ServerProxy>.__dir__ will be handled as xmlrpc call,
                                # expected behaviour is that the call returns list of strings for the scope
                                variables = []
                            for obj_key in variables:
                                try:
                                    obj_value = getattr(value, obj_key)
                                except:
                                    result.append("%20s - uninitialized variable" % ("self." + self._to_str(obj_key)))
                                    continue
                                if obj_key.startswith("__"):
                                    continue
                                if callable(obj_value):
                                    continue
                                obj_value_str = self._to_str(obj_value, "%.200r")
                                result.append("%20s = %s" % ("self." + self._to_str(obj_key), obj_value_str))
                    result.append("</LOCALS>")
        s = u''

        for i in result:
            line = i.replace(r'\\n', '\n').replace(r'\n', '\n')

            if type(i) == unicode:
                s += i
            else:
                s += unicode(str(i), errors='replace')
            s += '\n'

        return s.encode('ascii', 'replace')

    def print_traceback(self):
        """Print a traceback string to stderr."""
        print >>sys.stderr, self.get_traceback()

    def _get_lines_from_file(self, filename, lineno, context_lines):
        # this function was taken from Django and adapted for CLI
        """
        Return context_lines before and after lineno from file.
        Return (pre_context_lineno, pre_context, context_line, post_context).
        """
        source = None
        try:
            f = open(filename)
            try:
                source = f.readlines()
            finally:
                f.close()
        except (OSError, IOError):
            pass

        if source is None:
            return None, [], None, []

        encoding = "ascii"
        for line in source[:2]:
            # File coding may be specified. Match pattern from PEP-263
            # (http://www.python.org/dev/peps/pep-0263/)
            match = re.search(r"coding[:=]\s*([-\w.]+)", line)
            if match:
                encoding = match.group(1)
                break
        source = [ unicode(sline, encoding, "replace") for sline in source ]

        lower_bound = max(0, lineno - context_lines)
        upper_bound = lineno + context_lines

        pre_context = [ line.strip("\n") for line in source[lower_bound:lineno] ]
        context_line = source[lineno].strip("\n")
        post_context = [ line.strip("\n") for line in source[lineno+1:upper_bound] ]

        return lower_bound, pre_context, context_line, post_context


    def get_frames(self):
        # this function was taken from Django and adapted for CLI
        """Return a list with frame information."""
        tb = self.exc_info[2]

        frames = []
        while tb is not None:
            # support for __traceback_hide__ which is used by a few libraries to hide internal frames.
            if tb.tb_frame.f_locals.get("__traceback_hide__"):
                tb = tb.tb_next
                continue

            filename = tb.tb_frame.f_code.co_filename
            function = tb.tb_frame.f_code.co_name
            lineno = tb.tb_lineno - 1
            pre_context_lineno, pre_context, context_line, post_context = self._get_lines_from_file(filename, lineno, 7)
            if pre_context_lineno is not None:
                frames.append({
                    "tb": tb,
                    "filename": filename,
                    "function": function,
                    "lineno": lineno + 1,
                    "vars": tb.tb_frame.f_locals.items(),
                    "id": id(tb),
                    "pre_context": pre_context,
                    "context_line": context_line,
                    "post_context": post_context,
                    "pre_context_lineno": pre_context_lineno + 1,
                })
            tb = tb.tb_next

        if not frames:
            frames = [{
                "filename": "?",
                "function": "?",
                "lineno": "?",
            }]
        return frames


def set_except_hook(logger=None):
    """Replace standard excepthook method by an improved one."""
    def _hook(exctype, value, tb):
        tback = Traceback((exctype, value, tb))
        tback.show_locals = True
        logger and logger.error(tback.get_traceback())
        tback.print_traceback()
        print

    _hook.__doc__ = sys.excepthook.__doc__
    _hook.__name__ = sys.excepthook.__name__
    sys.excepthook = _hook



#!/usr/bin/python
# -*- coding: utf-8 -*-


import unittest2 as unittest
import run_tests # set sys.path

import os
import warnings
import tempfile
import shutil
import hashlib
import cPickle as pickle

# tolerate and skip in absence of rpm since it's not installable to virtualenv
try:
    import rpm
except ImportError:
    HAVE_RPM = False
else:
    from kobo.pkgset import *
    HAVE_RPM = True


@unittest.skipUnless(HAVE_RPM, "rpm python module is not installed")
class TestFileWrapperClass(unittest.TestCase):
    def setUp(self):
        self.tmp_dir = tempfile.mkdtemp()
        self.file_path = os.path.join(self.tmp_dir, "file")

    def tearDown(self):
        shutil.rmtree(self.tmp_dir)

    def test_file_name_property(self):
        open(self.file_path, "w").write("hello\n")
        wrap = FileWrapper(self.file_path)
        self.assertEqual(wrap.file_path, self.file_path)
        self.assertEqual(wrap.file_name, os.path.basename(self.file_path))

    def test_compute_checksums(self):
        open(self.file_path, "w").write("hello\n")

        res_origin = {}
        for name in ("md5", "sha1", "sha256", "sha512"):
            m = hashlib.new(name)
            m.update(open(self.file_path, "rb").read())
            res_origin[name] = m.hexdigest()

        wrap = FileWrapper(self.file_path)
        res = wrap.compute_checksums(["md5", "sha1", "sha256", "sha512"])
        self.assertEqual(res_origin, res)

    def test_pickle(self):
        self.test_file_name_property()
        name = "file"
        file1 = os.path.join(self.tmp_dir, name)
        wrap = FileWrapper(file1)
        pickled_data = pickle.dumps(wrap)
        wrap2 = pickle.loads(pickled_data)
        print wrap2.file_path


@unittest.skipUnless(HAVE_RPM, "rpm python module is not installed")
class TestRpmWrapperClass(unittest.TestCase):
    def setUp(self):
        self.file_path = "data/dummy-basesystem-10.0-6.noarch.rpm"
        self.source_path = "data/dummy-basesystem-10.0-6.src.rpm"
        self.nosource_path = "data/dummy-AdobeReader_enu-9.5.1-1.nosrc.rpm"

    def test_is_source(self):
        wrap = RpmWrapper(self.file_path)
        self.assertEqual(wrap.sourcepackage, None)
        self.assertEqual(wrap.is_source, False)
        self.assertEqual(wrap.nosource, [])
        wrap = RpmWrapper(self.source_path)
        self.assertEqual(wrap.sourcepackage, True)
        self.assertEqual(wrap.is_source, True)
        self.assertEqual(wrap.nosource, [])
        wrap = RpmWrapper(self.nosource_path)
        self.assertEqual(wrap.sourcepackage, True)
        self.assertEqual(wrap.is_source, True)
        self.assertEqual(wrap.nosource, [0])

    def test_is_system_release(self):
        wrap = RpmWrapper(self.file_path)
        self.assertEqual(wrap.is_system_release, False)

    def test_pickle(self):
        rpm_version = [int(v) for v in getattr(rpm, '__version__', '0.0').split('.')]
        if rpm_version[0:2] < [4, 10]:
            warnings.warn('RPM header pickling unsupported in rpm %s' % rpm_version)
            return
        wrap = RpmWrapper(self.file_path)
        pickled_data = pickle.dumps(wrap)
        wrap2 = pickle.loads(pickled_data)
        self.assertEqual(wrap.name, wrap2.name)


@unittest.skipUnless(HAVE_RPM, "rpm python module is not installed")
class TestSimpleRpmWrapperClass(unittest.TestCase):
    def setUp(self):
        self.file_path = "data/dummy-basesystem-10.0-6.noarch.rpm"
        self.source_path = "data/dummy-basesystem-10.0-6.src.rpm"

    def test_is_source(self):
        wrap = SimpleRpmWrapper(self.file_path)
        self.assertEqual(wrap.is_source, False)
        wrap = SimpleRpmWrapper(self.source_path)
        self.assertEqual(wrap.is_source, True)

    def test_is_system_release(self):
        wrap = RpmWrapper(self.file_path)
        self.assertEqual(wrap.is_system_release, False)

    def test_pickle(self):
        wrap = SimpleRpmWrapper(self.file_path)
        pickled_data = pickle.dumps(wrap)
        wrap2 = pickle.loads(pickled_data)
        self.assertEqual(wrap.name, wrap2.name)


@unittest.skipUnless(HAVE_RPM, "rpm python module is not installed")
class TestFileCacheClass(unittest.TestCase):
    def setUp(self):
        self.tmp_dir = tempfile.mkdtemp()
        self.cache = FileCache()
        self.file1 = os.path.join(self.tmp_dir, "file_1")
        self.file2 = os.path.join(self.tmp_dir, "file_2")

    def tearDown(self):
        shutil.rmtree(self.tmp_dir)

    def test_add_two_same_hardlinks(self):
        open(self.file1, "w").write("hello\n")
        os.link(self.file1, self.file2)

        self.cache = FileCache()
        wrap1 = self.cache.add(self.file1)
        wrap2 = self.cache.add(self.file2)

        self.assertEqual(len(self.cache.inode_cache), 1)
        self.assertEqual(len(self.cache.file_cache), 1)
        self.assertEqual(len(self.cache), 1)
        self.assertEqual(id(wrap1), id(wrap2))

    def test_add_two_different_files(self):
        open(self.file1, "w").write("roses are red\n")
        open(self.file2, "w").write("violets are blue\n")

        self.cache = FileCache()
        wrap1 = self.cache.add(self.file1)
        wrap2 = self.cache.add(self.file2)

        self.assertEqual(len(self.cache.inode_cache), 2)
        self.assertEqual(len(self.cache.file_cache), 2)
        self.assertEqual(len(self.cache), 2)
        self.assertNotEqual(id(wrap1), id(wrap2))

    def test_getitem(self):
        open(self.file1, "w").write("hello\n")
        open(self.file2, "w").write("hello\n")

        self.cache = FileCache()
        wrap1 = self.cache.add(self.file1)
        wrap2 = self.cache.add(self.file2)

        self.assertEqual(len(self.cache.inode_cache), 2)
        self.assertEqual(len(self.cache.file_cache), 2)
        self.assertEqual(id(self.cache[self.file1]), id(wrap1))
        self.assertEqual(id(self.cache[self.file2]), id(wrap2))

    def test_setitem(self):
        open(self.file1, "w").write("hello\n")

        self.cache1 = FileCache()
        self.cache2 = FileCache()
        wrap = self.cache1.add(self.file1)

        self.cache2[self.file1] = wrap

        self.assertEqual(len(self.cache2.file_cache), 1)
        self.assertEqual(id(self.cache2[self.file1]), id(wrap))

    def test_iteritems(self):
        open(self.file1, "w").write("hello\n")
        open(self.file2, "w").write("hello\n")

        self.cache = FileCache()
        wrap1 = self.cache.add(self.file1)
        wrap2 = self.cache.add(self.file2)

        items = [path for path, _ in self.cache.iteritems()]

        self.assertEqual(len(self.cache.inode_cache), 2)
        self.assertEqual(len(self.cache.file_cache), 2)
        self.assertEqual(len(items), 2)
        self.assertTrue(self.file1 in items)
        self.assertTrue(self.file2 in items)

    def test_iter(self):
        open(self.file1, "w").write("hello\n")
        open(self.file2, "w").write("hello\n")

        self.cache = FileCache()
        wrap1 = self.cache.add(self.file1)
        wrap2 = self.cache.add(self.file2)

        items = [item for item in self.cache]

        self.assertEqual(len(self.cache.inode_cache), 2)
        self.assertEqual(len(self.cache.file_cache), 2)
        self.assertEqual(len(items), 2)
        self.assertTrue(self.file1 in items)
        self.assertTrue(self.file2 in items)

    def test_remove_by_file_path(self):
        self.test_add_two_different_files()
        self.cache.remove(self.file1)

        items = [item for item in self.cache]

        self.assertEqual(len(self.cache.inode_cache), 1)
        self.assertEqual(len(self.cache.file_cache), 1)
        self.assertEqual(len(items), 1)
        self.assertTrue(self.file1 not in items)
        self.assertTrue(self.file2 in items)

    def test_remove_by_obj(self):
        self.test_add_two_different_files()

        self.file1_obj = self.cache[self.file1]
        self.cache.remove(self.file1_obj)

        items = [item for item in self.cache]

        self.assertEqual(len(self.cache.inode_cache), 1)
        self.assertEqual(len(self.cache.file_cache), 1)
        self.assertEqual(len(items), 1)
        self.assertTrue(self.file1 not in items)
        self.assertTrue(self.file2 in items)

    def test_remove_by_filenames(self):
        self.test_add_two_different_files()

        # add a file with existing name to a subdir
        os.makedirs(os.path.join(self.tmp_dir, "dir"))
        file1a = os.path.join(self.tmp_dir, "dir", "file_1")
        open(file1a, "w").write("hello\n")
        self.cache.add(file1a)

        self.cache.remove_by_filenames("does-not-exist")
        self.assertEqual(len(self.cache.inode_cache), 3)
        self.assertEqual(len(self.cache.file_cache), 3)

        # ignores the path, only the file name is important
        # removes both files with the file name "file_1"
        self.cache.remove_by_filenames("/foo/bar/file_1")

        items = [item for item in self.cache]

        self.assertEqual(len(self.cache.inode_cache), 1)
        self.assertEqual(len(self.cache.file_cache), 1)
        self.assertEqual(len(items), 1)
        self.assertTrue(self.file1 not in items)
        self.assertTrue(self.file2 in items)


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-


import django.contrib.admin as admin

from models import FileUpload


class FileUploadAdmin(admin.ModelAdmin):
    list_display = ('id', 'owner', 'name', 'size', 'upload_key', 'state', 'dt_created', 'dt_finished')
    list_filter = ('owner', 'state')
    search_fields = ('id', 'upload_key', 'name', 'dt_created', 'dt_finished')

admin.site.register(FileUpload, FileUploadAdmin)



#!/usr/bin/python
# -*- coding: utf-8 -*-

import unittest
import run_tests # set sys.path

# Following lines fix: django.core.exceptions.ImproperlyConfigured: The SECRET_KEY setting must not be empty.
import os
import sys
import imp
module_name = "test_types_settings"
settings_module = imp.new_module(module_name)
settings_module.SECRET_KEY = "foo"
sys.modules[module_name] = settings_module
os.environ["DJANGO_SETTINGS_MODULE"] = module_name

from kobo.types import *
from kobo.django.fields import StateEnumField


class TestEnum(unittest.TestCase):
    def setUp(self):
        self.enum = Enum(
            "A",
            "B",
            EnumItem("C", help_text="help", foo="foo", bar="bar"),
        )

    def test_duplicate_items(self):
        self.assertRaises(ValueError, Enum, "A", "A")
        self.assertRaises(ValueError, Enum, EnumItem("A"), "A")

    def test_in_enum(self):
        self.assert_("A" in self.enum)
        self.assert_("B" in self.enum)
        self.assert_("C" in self.enum)
        self.assert_("X" not in self.enum)

    def test_iter(self):
        self.assertEqual(list(self.enum), ["A", "B", "C"])

    def test_getitem(self):
        self.assertEqual(self.enum["A"], 0)
        self.assertEqual(self.enum["B"], 1)
        self.assertEqual(self.enum["C"], 2)

        self.assertEqual(self.enum[0], "A")
        self.assertEqual(self.enum[1], "B")
        self.assertEqual(self.enum[2], "C")

    def test_get(self):
        self.assertEqual(self.enum.get("B"), 1)
        self.assertEqual(self.enum.get(1), "B")

    def test_get_mapping(self):
        self.assertEqual(self.enum.get_mapping(), [(0, "A"), (1, "B"), (2, "C")])
        self.assertEqual(type(self.enum.get_mapping()[0][1]), str)

    def test_get_item(self):
        self.assertEqual(self.enum.get_item("A"), "A")
        self.assertRaises(KeyError, self.enum.get_item, "X")

    def test_get_help_text(self):
        self.assertEqual(self.enum.get_item_help_text("A"), "")
        self.assertEqual(self.enum.get_item_help_text("C"), "help")

    def test_get_option(self):
        self.assertEqual(self.enum.get_item_option("A", "foo"), None)
        self.assertEqual(self.enum.get_item_option("C", "foo"), "foo")

        self.assertEqual(self.enum.get_item_option("C", "bar"), "bar")
        self.assertEqual(self.enum.get_item_option("C", "bar", "default"), "bar")

        self.assertEqual(self.enum.get_item_option("C", "baz"), None)
        self.assertEqual(self.enum.get_item_option("C", "baz", "default"), "default")

        self.assertEqual(self.enum.get_item("C")["foo"], "foo")
        self.assertRaises(KeyError, self.enum.get_item("C").__getitem__, "baz")


class TestStateEnum(unittest.TestCase):
    def setUp(self):
        def enter_NEW(old_state, new_state, **kwargs):
            if str(new_state) != "NEW":
                raise RuntimeError("new_state == 'NEW' expected.")

        def leave_NEW(old_state, new_state, **kwargs):
            if str(old_state) != "NEW":
                raise RuntimeError("old_state == 'NEW' expected.")

        def enter_CLOSED(old_state, new_state, **kwargs):
            if str(new_state) != "CLOSED":
                raise RuntimeError("new_state == 'CLOSED' expected.")

        self.state_enum = StateEnum(
            State(
                name="NEW",
                next_states=["ASSIGNED", "MODIFIED"],
                enter=[enter_NEW],
                leave=[leave_NEW],
                check_perms=[],
            ),
            State(
                name="ASSIGNED",
                next_states=["MODIFIED"]
            ),
            State(
                name="MODIFIED",
                next_states=["ON_QA"]
            ),
            State(
                name="ON_QA",
                next_states=["VERIFIED", "FAILS_QA"]
            ),
            State(
                name="FAILS_QA",
                next_states=["ASSIGNED", "MODIFIED"]
            ),
            State(
                name="VERIFIED",
                next_states=["CLOSED"]),
            State(
                name="CLOSED",
                next_states=None,
                enter=[enter_CLOSED],
            ),
        )
        self.state_enum.set_state("NEW")

    def test_invalid_states(self):
        self.assertRaises(ValueError, StateEnum,
            State(
                name="NEW",
                next_states=["ASSIGNED", "MODIFIED"],
            ),
            State(
                name="ASSIGNED",
                next_states=["MODIFIED"]
            ),
        )

    def test_transitions(self):
        self.state_enum.set_state("NEW")
        self.state_enum.change_state("MODIFIED")
        self.state_enum.change_state("ON_QA")
        self.state_enum.change_state("VERIFIED")
        self.state_enum.change_state("CLOSED")

        self.state_enum.set_state("NEW")
        # don't commit the state transition -> False
        self.assertEqual(self.state_enum.change_state("MODIFIED", commit=False), False)
        self.assertEqual(self.state_enum.change_state("MODIFIED", commit=True), True)
        # stay in MODIFIED -> False
        self.assertEqual(self.state_enum.change_state("MODIFIED", commit=True), False)

        # and now without the commit argument
        self.state_enum.set_state("NEW")
        self.assertEqual(self.state_enum.change_state("MODIFIED"), True)
        self.assertEqual(self.state_enum.change_state("MODIFIED"), False)

    def test_invalid_transitions(self):
        self.assertRaises(ValueError, self.state_enum.change_state, "CLOSED")
        self.state_enum.change_state("MODIFIED")
        self.assertRaises(ValueError, self.state_enum.change_state, "NEW")
        self.assertRaises(ValueError, self.state_enum.change_state, 0)

    def test_final_states(self):
        self.assertEqual(self.state_enum.get_final_states(), ["CLOSED"])
        self.assertEqual(self.state_enum.get_final_states(return_id_list=True), [6])

    def test_commit(self):
        self.state_enum.set_state('NEW')
        self.assertEqual(self.state_enum._to, None)
        self.state_enum.change_state('MODIFIED', commit=False)
        self.assertEqual(str(self.state_enum), '0')
        self.state_enum.change_state('MODIFIED', commit=True)
        self.assertEqual(str(self.state_enum), '2')
        self.state_enum.change_state('ON_QA', commit=False)
        self.assertEqual(str(self.state_enum), '2')
        self.state_enum.change_state(None, commit=True)
        self.assertEqual(str(self.state_enum), '3')
        # no prepared new_state raises exception
        self.assertRaises(ValueError, self.state_enum.change_state, None, commit=True)


class TestStateEnumField(unittest.TestCase):
    def setUp(self):

        self.state_enum = StateEnum(
            State(
                name="NEW",
                next_states=["ASSIGNED", "MODIFIED"],
                check_perms=[],
            ),
            State(
                name="ASSIGNED",
                next_states=["MODIFIED"]
            ),
            State(
                name="MODIFIED",
                next_states=[]
            ),
        )
        self.state_enum.set_state("NEW")

        self.field = StateEnumField(self.state_enum, default='NEW')

    def test_to_python(self):
        t = self.field.to_python('0')
        self.assertEqual(type(t), StateEnum)
        self.assertEqual(t._current_state, 'NEW')
        t = self.field.to_python('1')
        self.assertEqual(t._current_state, 'ASSIGNED')
        t = self.field.to_python('NEW')
        self.assertEqual(type(t), StateEnum)
        self.assertEqual(t._current_state, 'NEW')
        t = self.field.to_python(t)
        self.assertEqual(type(t), StateEnum)
        self.assertEqual(t._current_state, 'NEW')
        t = self.field.to_python(2)
        self.assertEqual(type(t), StateEnum)
        self.assertEqual(t._current_state, 'MODIFIED')

    def test_choices(self):
        correct = ((0, 'NEW'), (1, 'ASSIGNED'), (2, 'MODIFIED'))
        self.assertEqual(correct, self.field.choices)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-


from django.core.urlresolvers import reverse
from django.core.exceptions import ObjectDoesNotExist

import kobo.hub.models as models
from kobo.django.xmlrpc.decorators import admin_required, login_required


__all__ = (
    "enable_worker",
    "disable_worker",
    "get_worker_info",
    "shutdown_worker",
    "task_info",
    "get_tasks",
    "cancel_task",
    "resubmit_task",
    "list_workers",
    "create_task",
    "task_url",
)


@admin_required
def shutdown_worker(request, worker_name, kill=False):
    """shutdown_worker(worker_name, kill): None

    Send shutdown request to a worker.
    If kill argument is set to True, kill worker immediately,
    otherwise wait until all tasks assigned to the are finished.
    """
    return models.Task.create_shutdown_task(request.user.username, worker_name, kill=kill)


@admin_required
def enable_worker(request, worker_name):
    """enable_worker(worker_name): none
    """
    models.Worker.objects.filter(name = worker_name).update(enabled=True)


@admin_required
def disable_worker(request, worker_name):
    """disable_worker(worker_name, kill): None
    """
    models.Worker.objects.filter(name = worker_name).update(enabled=False)

@admin_required
def get_worker_info(request, worker_name):
    try:
        return models.Worker.objects.get(name = worker_name).export()
    except models.Worker.DoesNotExist:
        return {}

def task_info(request, task_id, flat=False):
    """task_info(task_id, flat=False): dict or None"""
    task = models.Task.objects.get(id=task_id)
    return task.export(flat=flat)


def get_tasks(request, task_id_list, state_list=None):
    """get_tasks(task_id_list): list

    @param task_id_list: list of task ids, can be empty, then all tasks are
    retrieved
    @type task_id_list: [int]
    @param state_list: task state ids by which task_id_list should be
    filtered
    @type: [int]
    @return: list of task_info dicts
    @rtype: list
    """

    if task_id_list:
        tasks = models.Task.objects.filter(id__in=task_id_list)
    else:
        tasks = models.Task.objects.all()
    if state_list:
        tasks = tasks.filter(state__in=state_list)
    return [ i.export(flat=True) for i in tasks ]


@login_required
def cancel_task(request, task_id):
    try:
        task = models.Task.objects.get(id=task_id)
    except ObjectDoesNotExist:
        return "Specified task %s does not exist." % task_id
    return task.cancel_task(user=request.user)


@login_required
def resubmit_task(request, task_id, force=False):
    """Resubmit a failed task and return new task_id."""
    task = models.Task.objects.get(id=task_id)
    return task.resubmit_task(request.user, force)


def list_workers(request, enabled=True):
    """(): [string]"""
    return sorted([ worker.name for worker in models.Worker.objects.filter(enabled=enabled) ])


@admin_required
def create_task(request, kwargs):
    """
    Create a new task which is either brand new or based on existing one.
    This call can be invoked only by superuser.

    @param kwargs: task attributes: owner_name, label, method, args, comment, parent_id, worker_name, arch_name, channel_name, timeout, priority, weight, exclusive
                   when task_id is set, the task is used as a template for the new one
    @type kwargs: dict
    @return: task id
    @rtype: int
    """

    old_task_id = kwargs.pop("task_id", None)
    if old_task_id is not None:
        old_task = models.Task.objects.get(id=old_task_id)
        return old_task.clone_task(request.user, **kwargs)

    kwargs.setdefault("label", "")
    kwargs["resubmitted_by"] = request.user
    kwargs["resubmitted_from"] = None
    return models.Task.create_task(**kwargs)


def task_url(request, task_id):
    """
    Get a task URL.

    @param task_id: task ID
    @type  task_id: int
    @return: task URL
    @rtype:  str
    """
    prefix = request.META["SERVER_PORT"] == 443 and "https://" or "http://"

    # use HTTP_HOST (address can point to a virtual host)
    # if address points to localhost, use SERVER_NAME in order to make link globally valid
    server_name = request.META["HTTP_HOST"]
    if server_name in ("localhost", "localhost.localdomain"):
        server_name = request.META["SERVER_NAME"]

    return prefix + server_name + reverse("task/detail", args=[task_id])



import re
from django.db import models
from django.core import validators
from django.utils.translation import ugettext_lazy as _
from django.contrib.auth.models import AbstractBaseUser, PermissionsMixin, UserManager
from django.core.mail import send_mail
from django.utils import timezone

MAX_LENGTH = 255

class User(AbstractBaseUser, PermissionsMixin):
    """
    Copy (non-abstract) of AbstractUser with longer username. Removed profile support as it
    is deprecated in django 1.5.

    Username, password and email are required. Other fields are optional.
    """
    username = models.CharField(_('username'), max_length=MAX_LENGTH, unique=True,
        help_text=_('Required. %s characters or fewer. Letters, digits and @/./+/-/_ only.' % MAX_LENGTH),
        validators=[
            validators.RegexValidator(r'^[\w.@+-]+$',
                                      _('Enter a valid username. '
                                        'This value may contain only letters, numbers '
                                        'and @/./+/-/_ characters.'), 'invalid'),
        ],
        error_messages={
            'unique': _("A user with that username already exists."),
        })
    first_name = models.CharField(_('first name'), max_length=30, blank=True)
    last_name = models.CharField(_('last name'), max_length=30, blank=True)
    email = models.EmailField(_('email address'), blank=True)
    is_staff = models.BooleanField(_('staff status'), default=False,
        help_text=_('Designates whether the user can log into this admin '
                    'site.'))
    is_active = models.BooleanField(_('active'), default=True,
        help_text=_('Designates whether this user should be treated as '
                    'active. Unselect this instead of deleting accounts.'))
    date_joined = models.DateTimeField(_('date joined'), default=timezone.now)

    objects = UserManager()

    USERNAME_FIELD = 'username'
    REQUIRED_FIELDS = ['email']

    class Meta:
        db_table = 'auth_user'
        verbose_name = _('user')
        verbose_name_plural = _('users')

    def get_full_name(self):
        """
        Returns the first_name plus the last_name, with a space in between.
        """
        full_name = '%s %s' % (self.first_name, self.last_name)
        return full_name.strip()

    def get_short_name(self):
        "Returns the short name for the user."
        return self.first_name

    def email_user(self, subject, message, from_email=None, **kwargs):
        """
        Sends an email to this User.
        """
        send_mail(subject, message, from_email, [self.email], **kwargs)



# -*- coding: utf-8 -*-


import sys
from xmlrpclib import Fault

from kobo.client import ClientCommand


class Shutdown_Worker(ClientCommand):
    """shutdown a worker"""
    enabled = True
    admin = True


    def options(self):
        self.parser.usage = "%%prog %s [--kill] worker_name [worker_name]" % self.normalized_name

        self.parser.add_option(
            "--kill",
            default=False,
            action="store_true",
            help="kill worker immediately"
        )


    def run(self, *args, **kwargs):
        kill = kwargs.pop("kill", False)

        if len(args) == 0:
            self.parser.error("No worker specified.")

        username = kwargs.pop("username", None)
        password = kwargs.pop("password", None)
        workers = args

        self.set_hub(username, password)
        for worker in workers:
            try:
                self.hub.client.shutdown_worker(worker, kill)
            except Fault, ex:
                sys.stderr.write("%s\n" % ex.faultString)



# -*- coding: utf-8 -*-


"""
CommandOptionParser HOWTO
=========================

1) setup CommandContainer
-------------------------
# In case you don't need any special functionality, just import default CommandContainer.
# Otherwise it's recommended to inherit your own container and extend it's functionality.
# Typical use cases are shared configuration or shared XML-RPC connection.

from kobo.cli import CommandContainer
class MyCommandContainer(CommandContainer):
    def __init__(self, *args, **kwargs):
        CommandContainer.__init__(self, *args, **kwargs)
        self.xmlrpc_client = ...


2) write your own Commands
--------------------------
# It usually makes sense to inherit directly from Command class.
# All common methods and attributes should be in the container.
# Specify any OptionParser options in options() method.
# OptionParser.parse_args() result is automatically passed to run(*args, **kwargs) method.
# A OptionParser instance os available in self.parser attribute.

class Make_Dirs(Command):
    '''create directories'''
    enabled = True
    admin = False

    def options(self):
        self.parser.usage = "%%prog %s [options] <user>" % self.normalized_name
        self.parser.add_option("-m", "--mode", help="set directory perms (0xxx)")

    def run(self, *args, **kwargs):
        if len(args) < 1:
            self.parser.error("Please specify a directory")
        mode = kwargs.pop("mode", "0755")
        mode = int(mode, 0) # convert oct string to int

        import os
        for directory in args:
            os.makedirs(directory, mode=mode)


3) register commands to a container
-----------------------------------
# Register either either all plugins (register_plugin)
# or all plugins in a module (register_module) to a container.
# All plugins must have enabled=True otherwise they won't be registered.

(My)CommandContainer.register_plugin(plugin_class)
(My)CommandContainer.register_module(module_with_plugins)


4) Use CommandOptionParser
--------------------------
command_container = (My)CommandContainer()
parser = CommandOptionParser(command_container=command_container)
parser.run()

# See kobo.client.main for slightly advanced example.
"""


import sys
import optparse
import datetime
from optparse import Option
from xmlrpclib import Fault

from kobo.plugins import Plugin, PluginContainer
from kobo.shortcuts import force_list


__all__ = (
    "Command",
    "CommandContainer",
    "CommandOptionParser",
    "Option",
    "username_prompt",
    "password_prompt",
    "yes_no_prompt",
    "are_you_sure_prompt",
)


def username_prompt(prompt=None, default_value=None):
    """Ask for a username."""
    if default_value is not None:
        return default_value

    prompt = prompt or "Enter your username: "
    print >>sys.stderr, prompt,
    return sys.stdin.readline()


def password_prompt(prompt=None, default_value=None):
    """Ask for a password."""
    import getpass

    if default_value is not None:
        return default_value

    prompt = prompt or "Enter your password: "
    try:
        # try to use stderr stream
        result = getpass.getpass(prompt, stream=sys.stderr)
    except TypeError:
        # fall back to stdout
        result = getpass.getpass(prompt)
    return result


def yes_no_prompt(prompt, default_value=None):
    """Give a yes/no (y/n) question."""
    if default_value is not None:
        if default_value not in ("Y", "N"):
            raise ValueError("Invalid default value: %s" % default_value)
        default_value = default_value.upper()

    prompt = "%s [%s/%s]: " % (prompt, ("y", "Y")[default_value == "Y"], ("n", "N")[default_value == "N"])
    print >>sys.stderr, prompt,

    while True:
        user_input = sys.stdin.readline().strip().upper()
        if user_input == "" and default_value is not None:
            user_input = default_value

        if user_input == "Y":
            return True
        if user_input == "N":
            return False


def are_you_sure_prompt(prompt=None):
    """Give a yes/no (y/n) question."""
    prompt = prompt or "Are you sure? Enter 'YES' to continue: "
    print >>sys.stderr, prompt,
    user_input = sys.stdin.readline().strip()

    if user_input == "YES":
        return True

    return False


class Command(Plugin):
    """An abstract class representing a command for CommandOptionParser."""

    enabled = False
    admin = False

    username_prompt = staticmethod(username_prompt)
    password_prompt = staticmethod(password_prompt)
    yes_no_prompt = staticmethod(yes_no_prompt)
    are_you_sure_prompt = staticmethod(are_you_sure_prompt)

    def __init__(self, parser):
        Plugin.__init__(self)
        self.parser = parser

    def options(self):
        """Add options to self.parser."""
        pass

    def run(self, *args, **kwargs):
        """Run a command. Arguments contain parsed options."""
        raise NotImplementedError()


class CommandContainer(PluginContainer):
    """Container for Command classes."""

    @classmethod
    def normalize_name(cls, name):
        """Replace some characters in command names."""
        return name.lower().replace('_', '-').replace(' ', '-')


class CommandOptionParser(optparse.OptionParser):
    """Enhanced OptionParser with plugin support."""
    def __init__(self,
            usage=None,
            option_list=None,
            option_class=Option,
            version=None,
            conflict_handler="error",
            description=None,
            formatter=None,
            add_help_option=True,
            prog=None,
            command_container=None,
            default_command="help",
            add_username_password_options=False):

        usage = usage or "%prog <command> [args] [--help]"
        self.container = command_container
        self.default_command = default_command
        self.command = None
        formatter = formatter or optparse.IndentedHelpFormatter(max_help_position=33)

        optparse.OptionParser.__init__(self, usage, option_list, option_class, version, conflict_handler, description, formatter, add_help_option, prog)

        if add_username_password_options:
            option_list = [
                optparse.Option("--username", help="specify user"),
                optparse.Option("--password", help="specify password"),
            ]
            self._populate_option_list(option_list, add_help=False)

    def print_help(self, file=None, admin=False):
        if file is None:
            file = sys.stdout
        file.write(self.format_help())
        if self.command in (None, "help", "help-admin"):
            file.write("\n")
            file.write(self.format_help_commands(admin=admin))

    def format_help_commands(self, admin=False):
        commands = []
        admin_commands = []

        for name, plugin in sorted(self.container.plugins.iteritems()):
            is_admin = getattr(plugin, "admin", False)
            text = "  %-30s %s" % (name, plugin.__doc__ or "")
            if is_admin:
                if admin:
                    admin_commands.append(text)
            else:
                commands.append(text)

        if commands:
            commands.insert(0, "commands:")
            commands.append("")

        if admin_commands:
            admin_commands.insert(0, "admin commands:")
            admin_commands.append("")

        return "\n".join(commands + admin_commands)

    def parse_args(self, args=None, values=None):
        """return (command_instance, opts, args)"""
        args = self._get_args(args)
        command = None

        if len(args) > 0 and not args[0].startswith("-"):
            command = args[0]
            args = args[1:]
        else:
            command = self.default_command
            # keep args as is

        if not command in self.container.plugins:
            self.error("unknown command: %s" % command)

        CommandClass = self.container[command]
        cmd = CommandClass(self)
        if self.command != cmd.normalized_name:
            self.command = cmd.normalized_name
            cmd.options()
        cmd_opts, cmd_args = optparse.OptionParser.parse_args(self, args, values)
        return (cmd, cmd_opts, cmd_args)

    def run(self, args=None, values=None):
        """parse arguments and run a command"""
        cmd, cmd_opts, cmd_args = self.parse_args(args, values)
        cmd_kwargs = cmd_opts.__dict__
        cmd.run(*cmd_args, **cmd_kwargs)


class Help(Command):
    """show this help message and exit"""
    enabled = True

    def options(self):
        pass

    def run(self, *args, **kwargs):
        self.parser.print_help(admin=False)


class Help_Admin(Command):
    """show help message about administrative commands and exit"""
    enabled = True

    def options(self):
        # override default --help option
        opt = self.parser.get_option("--help")
        opt.action = "store_true"
        opt.dest = "help"

    def run(self, *args, **kwargs):
        self.parser.print_help(admin=True)


class Help_RST(Command):
    """print program usage as reStructuredText"""
    enabled = True

    def options(self):
        pass

    def run(self, *args, **kwargs):
        prog = self.parser.get_prog_name()
        print ".. -*- coding: utf-8 -*-"
        print
        print "=" * len(prog)
        print prog
        print "=" * len(prog)
        print

        # add subtitle (command description)
        description = getattr(self.parser.container, "_description", None)
        if description:
            print ":Subtitle: %s" % description
            print

        # add copyright
        copyright = getattr(self.parser.container, "_copyright", None)
        if copyright:
            print ":Copyright: %s" % copyright

        # add date
        print ":Date: $Date: %s $" % datetime.datetime.strftime(datetime.datetime.utcnow(), format="%F %X")
        print

        print "--------"
        print "COMMANDS"
        print "--------"

        for command_name, CommandClass in sorted(self.parser.container.plugins.items()):
            parser = optparse.OptionParser(usage=self.parser.usage)
            cmd = CommandClass(parser)
            cmd.normalized_name = command_name
            cmd.options()
            cmd.container = self.parser.container
            cmd_opts, cmd_args = parser.parse_args()

            print command_name
            print "-" * len(command_name)

            if cmd.admin:
                print "[ADMIN ONLY]",

            print cmd.__doc__.strip()
            print
            usage = parser.get_usage().strip().replace("Usage: ", "**Usage:** ", 1)
            if usage:
                print usage
                print 

            for opt in sorted(parser.option_list, lambda x, y: cmp(str(x), str(y))):
                if "-h/--help" in str(opt):
                    continue
                if opt.nargs:
                    metavar = opt.metavar or opt.dest.upper()
                opt_list = []
                for opt_str in opt._short_opts + opt._long_opts:
                    if opt.nargs is not None:
                        opt_list.append("%s=%s" % (opt_str, metavar))
                    else:
                        opt_list.append(opt_str)
                print "/".join(opt_list)
                print "  %s" % opt.help
                if opt.action == "append":
                    print
                    print "  This option can be specified multiple times"
                print
            print

        # handle :Contact: and :Author: ourselves
        authors = force_list(getattr(self.parser.container, "_authors", []))
        contact = getattr(self.parser.container, "_contact", None)
        if authors or contact:
            print "-------"
            print "AUTHORS"
            print "-------"

            for author in sorted(authors):
                print "- %s" % author
            print

            if contact:
                print "**Contact:** %s" % contact
                print


CommandContainer.register_plugin(Help)
CommandContainer.register_plugin(Help_Admin)
CommandContainer.register_plugin(Help_RST)



from django.conf.urls import url, patterns, include

urlpatterns = patterns("",
    url(r"^auth/", include("kobo.hub.urls.auth")),
    url(r"^task/", include("kobo.hub.urls.task")),
    url(r"^info/arch/", include("kobo.hub.urls.arch")),
    url(r"^info/channel/", include("kobo.hub.urls.channel")),
    url(r"^info/user/", include("kobo.hub.urls.user")),
    url(r"^info/worker/", include("kobo.hub.urls.worker")),
)



# -*- coding: utf-8 -*-
from south.db import db
from south.v2 import SchemaMigration

class Migration(SchemaMigration):

    def forwards(self, orm):
        db.rename_column('auth_user_groups',           'longnameuser_id', 'user_id')
        db.rename_column('auth_user_user_permissions', 'longnameuser_id', 'user_id')

    def backwards(self, orm):
        db.rename_column('auth_user_groups',           'user_id', 'longnameuser_id')
        db.rename_column('auth_user_user_permissions', 'user_id', 'longnameuser_id')



# -*- coding: utf-8 -*-


import sys
from xmlrpclib import Fault

from kobo.client import ClientCommand


class Disable_Worker(ClientCommand):
    """disable worker"""
    enabled = True
    admin = True


    def options(self):
        self.parser.usage = "%%prog %s [--all] [worker_name]" % self.normalized_name

        self.parser.add_option(
            "--all",
            default=False,
            action="store_true",
            help="Disable all enabled workers"
        )

    def run(self, *args, **kwargs):
        if len(args) == 0 and not kwargs['all']:
            self.parser.error("No worker (or --all) specified.")
        if len(args) and kwargs['all']:
            self.parser.error("Specify worker name or --all. From safety reasons both are not allowed.")

        username = kwargs.pop("username", None)
        password = kwargs.pop("password", None)

        self.set_hub(username, password)
        if kwargs['all']:
            try:
                workers = self.hub.client.list_workers(True)
            except Fault, ex:
                sys.stderr.write("%s\n" % ex.faultString)
                sys.exit(1)
        else:
            workers = args
        for worker in workers:
            try:
                self.hub.client.disable_worker(worker)
            except Fault, ex:
                sys.stderr.write("%s\n" % ex.faultString)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import unittest2 as unittest
from kobo.hub.models import _utf8_chunk as utf8_chunk

class TestUtf8Chunk(unittest.TestCase):
    def test_noop_ascii(self):
        """utf8_chunk returns input bytes if byte sequence is entirely ASCII"""
        bytestr = b'hello world'
        self.assertIs(utf8_chunk(bytestr), bytestr)

    def test_noop_utf8_end(self):
        """utf8_chunk returns input bytes if byte sequence uses non-ASCII
        UTF-8 at the end of the string and is well-formed"""
        unistr = u'hello 世界'
        bytestr = unistr.encode('utf-8')
        self.assertIs(utf8_chunk(bytestr), bytestr)

    def test_noop_utf8_mid(self):
        """utf8_chunk returns input bytes if byte sequence uses non-ASCII
        UTF-8 in the middle of the string and is well-formed"""
        unistr = u'hello 世界!'
        bytestr = unistr.encode('utf-8')
        self.assertIs(utf8_chunk(bytestr), bytestr)

    def test_noop_invalid(self):
        """utf8_chunk returns input bytes if byte sequence is not valid
        UTF-8 and can't be fixed by truncation"""
        bytestr = b'hello \xff\xff\xff'
        self.assertIs(utf8_chunk(bytestr), bytestr)

    def test_fixup_end(self):
        """utf8_chunk returns copy of input aligned to nearest character boundary
        if input is a byte sequence truncated in the middle of a unicode character."""
        unistr = u'hello 世界'
        bytestr = unistr.encode('utf-8')

        # this is now a broken sequence since we cut it off
        # partway through a character
        bytestr = bytestr[:-1]

        # proving it's broken
        try_decode = lambda: bytestr.decode('utf-8')
        self.assertRaises(UnicodeDecodeError, try_decode)

        # utf8_chunk unbreaks it by removing until the previous
        # complete character
        self.assertEqual(utf8_chunk(bytestr).decode('utf-8'), u'hello 世')

if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-


from django.utils.translation import ugettext_lazy as _
from django.conf.urls import url, patterns
from kobo.django.views.generic import ExtraListView
from kobo.hub.views import DetailViewWithWorkers
from kobo.hub.models import Channel

urlpatterns = patterns("",
    url(r"^$", ExtraListView.as_view(
        queryset=Channel.objects.order_by("name"),
        template_name="channel/list.html",
        context_object_name="channel_list",
        title=_("Channels"),
    ), name="channel/list"),
    url(r"^(?P<pk>\d+)/$", DetailViewWithWorkers.as_view(
        model = Channel,
        template_name = "channel/detail.html",
        context_object_name = "channel",
        title = _("Architecture detail"),
    ), name="channel/detail"),
)



# -*- coding: utf-8 -*-


import copy
import signal

from kobo.plugins import Plugin
from kobo.shortcuts import force_list
from kobo.client.constants import TASK_STATES


__all__ = (
    "TaskBase",
    "FailTaskException",
)


class FailTaskException(Exception):
    """Terminate the task."""
    pass


class TaskBase(Plugin):
    """Base class for all tasks."""
    enabled = True

    def __init__(self, hub, conf, task_id, args):
        self._hub = hub            # created by taskmanager
        self._conf = conf
        self._task_id = task_id
        self._task_info = self.hub.worker.get_task(self.task_id)
        self._task_manager = None  # created by taskmanager (only for foreground tasks)
        self._args = args
        self._subtask_list = []
        self.result = ""

    @property
    def hub(self):
        return self._hub

    @property
    def conf(self):
        return self._conf

    @property
    def task_id(self):
        return self._task_id

    @property
    def task_info(self):
        return self._task_info

    def _get_task_manager(self):
        if not getattr(self.__class__, "foreground", False):
            return None
        return self._task_manager

    def _set_task_manager(self, value):
        if not getattr(self.__class__, "foreground", False):
            raise ValueError("Cannot set task_manager for a background task.")
        self._task_manager = value

    task_manager = property(_get_task_manager, _set_task_manager)

    @property
    def args(self):
        # deepcopy to prevent modification
        return copy.deepcopy(self._args)

    @property
    def subtask_list(self):
        # deepcopy to prevent modification
        return copy.deepcopy(self._subtask_list)

    def run(self):
        """Run the task."""
        raise NotImplementedError()

    def fail(self):
        """Fail the task."""
        raise FailTaskException()

    @classmethod
    def cleanup(cls, hub, conf, task_info):
        pass

    @classmethod
    def notification(cls, hub, conf, task_info):
        pass

    def spawn_subtask(self, method, args, label=""):
        """Spawn a new subtask."""
        if self.foreground:
            raise RuntimeError("Foreground tasks can't spawn subtasks.")

        subtask_id = self.hub.worker.create_subtask(label, method, args, self.task_id)
        self._subtask_list.append(subtask_id)
        return subtask_id

    def wait(self, subtasks=None):
        """Wait until subtasks finish.

        subtasks = None - wait for all subtasks
        subtasks = [task_id list] - wait for selected subtasks
        """

        if self.foreground:
            # wait would call signal.pause() in the *main* worker thread and lock program forever
            raise RuntimeError("Foreground tasks can't wait on subtasks.")

        if subtasks is not None:
            subtasks = force_list(subtasks)

        self.hub.worker.wait(self.task_id, subtasks)

        finished = []
        while True:
            (finished, unfinished) = self.hub.worker.check_wait(self.task_id)

            if len(unfinished) == 0:
                # all done
                break

            # sleep
            signal.pause()
            # wake up on signal to check the status

        # remove finished subtasks from the list, check results
        fail = False
        for i in finished:
            state = self.hub.worker.get_task(i)
            if state['state'] != TASK_STATES['CLOSED']:
                fail = True
            self._subtask_list.remove(i)

        if fail:
            print "Failing because of at least one subtask hasn't closed properly."
            self.fail()

        return finished



# -*- coding: utf-8 -*-

DATABASES = {
    'default': {
        'NAME': ':memory:',
        'ENGINE': 'django.db.backends.sqlite3'
    }
}

INSTALLED_APPS = (
    'fields_test',
)

SECRET_KEY = 'whatever'



# -*- coding: utf-8 -*-

import warnings
from django.conf import settings
from django.db.models.query import QuerySet
from django.views.generic.edit import ProcessFormView, FormMixin
from django.views.generic.list import ListView
from django.views.generic.detail import DetailView

class ExtraListView(ListView):
    paginate_by = getattr(settings, "PAGINATE_BY", None)
    extra_context = None
    title = None

    def get_context_data(self, **kwargs):
        context = super(ExtraListView, self).get_context_data(**kwargs)
        if self.extra_context is not None:
            context.update(self.extra_context)
        if self.title is not None:
            context['title'] = self.title
        return context

class ExtraDetailView(DetailView):
    extra_context = None
    title = None

    def get_context_data(self, **kwargs):
        context = super(ExtraDetailView, self).get_context_data(**kwargs)
        if self.extra_context is not None:
            context.update(self.extra_context)
        if self.title is not None:
            context['title'] = self.title
        return context

class SearchView(FormMixin, ProcessFormView, ExtraListView):
    """
    Appends GET variables to the context. Used for paginated searches.
    In settings.py you can set optional parameter PAGINATE_BY which will be
    used as a default value.

    Useful fields:
        form_class - form_class is expected that it implements get_query(request)
                     method which returns queryset for displaying in list. It
                     also can return just query (Q()) which is then applied
                     to base queryset/model of class
    """

    object_list = []

    # ProcessFormView needs redefine get to use form data
    def get(self,  request,  *args,  **kwargs):
        return self.post(request,  *args,  **kwargs)

    # same goes for FormMixin
    def get_form_kwargs(self):
        kwargs = {'initial': self.get_initial()}
        if self.request.method == 'GET':
            kwargs['data'] = self.request.GET
        return kwargs

    def form_invalid(self, form):
        self.queryset = []
        return self.render_to_response(self.get_context_data(form=form))

    def form_valid(self, form):
        query = form.get_query(self.request)
        if isinstance(query, QuerySet):
            self.queryset = query
        else:
            self.queryset = self.get_queryset().filter(query)
        return self.render_to_response(self.get_context_data(form=form))

    def get_context_data(self, **kwargs):
        kwargs['object_list'] = self.get_queryset()
        context = super(SearchView, self).get_context_data(**kwargs)

        get_vars = self.request.GET.copy()
        if "page" in get_vars:
            del get_vars["page"]
        if len(get_vars) > 0:
            get_vars = "&%s" % get_vars.urlencode()
        else:
            get_vars = ""
        context['get_vars'] = get_vars
        return context

def object_list(request, **kwargs):
    warnings.warn('object_list is deprecated, please use ExtraListView or SearchView in future.')
    if 'form' in kwargs:
        return SearchView.as_view(**kwargs)(request)
    else:
        return ExtraListView.as_view(**kwargs)(request)



# -*- coding: utf-8 -*-


import datetime
import base64
import socket

import django.contrib.auth
from django.conf import settings
from django.contrib.auth.backends import ModelBackend
from django.core.exceptions import PermissionDenied
from django.contrib.sessions.models import Session

from kobo.django.auth.krb5 import Krb5RemoteUserBackend


__all__ = (
    "renew_session",
    "login_krbv",
    "login_password",
    "logout",
)


def renew_session(request):
    """renew_session(): bool

    Renew current session. Return True if session is no longer valid and user should log in.
    """

    request.session.modified = True
    return not request.user.is_authenticated()


def login_password(request, username, password):
    """login_password(username, password): session_id"""
    backend = ModelBackend()
    user = backend.authenticate(username, password)
    if user is None:
        raise PermissionDenied("Invalid username or password.")
    user.backend = "%s.%s" % (backend.__module__, backend.__class__.__name__)
    django.contrib.auth.login(request, user)
    return request.session.session_key


# TODO: proxy_user
def login_krbv(request, krb_request, proxy_user=None):
    """login_krbv(krb_request, proxy_user=None): session_key"""
    import krbV

    context = krbV.default_context()
    server_principal = krbV.Principal(name=settings.KRB_AUTH_PRINCIPAL, context=context)
    server_keytab = krbV.Keytab(name=settings.KRB_AUTH_KEYTAB, context=context)

    auth_context = krbV.AuthContext(context=context)
    auth_context.flags = krbV.KRB5_AUTH_CONTEXT_DO_SEQUENCE | krbV.KRB5_AUTH_CONTEXT_DO_TIME
    auth_context.addrs = (socket.gethostbyname(request.META["HTTP_HOST"]), 0, request.META["REMOTE_ADDR"], 0)

    # decode and read the authentication request
    decoded_request = base64.decodestring(krb_request)
    auth_context, opts, server_principal, cache_credentials = context.rd_req(decoded_request, server=server_principal, keytab=server_keytab, auth_context=auth_context, options=krbV.AP_OPTS_MUTUAL_REQUIRED)
    cprinc = cache_credentials[2]

    # remove @REALM
    username = cprinc.name.split("@")[0]
    backend = Krb5RemoteUserBackend()
    user = backend.authenticate(username)
    if user is None:
        raise PermissionDenied()
    user.backend = "%s.%s" % (backend.__module__, backend.__class__.__name__)
    django.contrib.auth.login(request, user)
    return request.session.session_key


def logout(request):
    """logout()

    Delete session information.
    """
    django.contrib.auth.logout(request)



# -*- coding: utf-8 -*-


import sys
from xmlrpclib import Fault

from kobo.client import ClientCommand


class Add_User(ClientCommand):
    """add a new user"""
    enabled = True
    admin = True


    def options(self):
        self.parser.usage = "%%prog %s [options] <user>" % self.normalized_name

        self.parser.add_option(
            "--admin",
            default=False,
            action="store_true",
            help="grant admin privileges"
        )


    def run(self, *args, **kwargs):
        if len(args) < 1:
            self.parser.error("Please specify a user")

        username = kwargs.pop("username", None)
        password = kwargs.pop("password", None)
        admin = kwargs.pop("admin", False)
        user = args[0]

        self.set_hub(username, password)
        try:
            self.hub.admin.add_user(user, admin)
        except Exception, ex:
            print str(ex)
            sys.exit(1)



# -*- coding: utf-8 -*-

import os
import sys
import datetime
import gzip
import shutil
import logging
from collections import deque
import io

logger =  logging.getLogger("kobo")

try:
    import json
except ImportError:
    import simplejson as json

from django.conf import settings
from django.contrib.auth import get_user_model
from django.core.exceptions import MultipleObjectsReturned, ObjectDoesNotExist
from django.db import models, connection, transaction
from django.utils.translation import ugettext_lazy as _
from django.db.models.signals import post_delete

import kobo.django.fields
from kobo.client.constants import *
from kobo.shortcuts import random_string, read_from_file, save_to_file, run

LOG_BUFFER_SIZE = 2**20


def dump_dict(**kwargs):
    """Serialize args dictionary to a json dump."""
    return json.dumps(kwargs)


def load_dict(dump):
    """Deserialize a json dump to dictionary."""
    return json.loads(dump)


def _utf8_chunk(bytestr):
    """If bytestr is a valid UTF-8 sequence, returns bytestr.
    If bytestr can be made a valid UTF-8 sequence by removing a few bytes
    from the end, returns a modified copy of bytestr.
    Otherwise, returns bytestr.

    The purpose of this method is to provide a safe way of reading chunks
    from the middle of a UTF-8 log file given an offset and length.
    Consider a caller reading a log file in 1MB chunks.  It's possible that
    a chunk would end in the middle of a UTF-8 encoded character, resulting
    in a broken string.  This method fixes up any such broken chunks."""
    try:
        bytestr.decode('utf-8')
        # It's already fully valid
        return bytestr
    except UnicodeDecodeError:
        pass

    # bytestr is not valid.
    # Can we make it valid by chopping a few bytes from the end?
    idx = len(bytestr)
    while idx > 0:
        last_char = ord(bytestr[idx - 1])
        if last_char < 0x80:
            # ascii - safe
            break
        idx = idx - 1
        if last_char >= 0xC0:
            # first byte of multi-byte character - not safe,
            # but previous byte is safe
            break

    out = bytestr[:idx]
    try:
        out.decode('utf-8')
        # We could fix up the string
        return out
    except UnicodeDecodeError:
        # We could not fix it up - might be broken for other reasons.
        # Just return unmodified bytes, so any decoding error is raised
        # from the caller
        return bytestr


def _tail(fh, max_size, max_line_size):
    """See TaskLogs.tail"""
    buffer = deque()
    current_size = 0
    offset = 0

    while True:
        this_line = fh.readline(max_line_size)
        if not this_line:
            break

        this_len = len(this_line)

        current_size = current_size + this_len
        offset = offset + this_len

        buffer.append(this_line)
        while current_size > max_size:
            removed = buffer.popleft()
            current_size = current_size - len(removed)

    return b''.join(buffer), offset


class Arch(models.Model):
    """Model for hub_arch table."""
    name        = models.CharField(max_length=16, unique=True, help_text=_("i386, ia64, ..."))
    pretty_name = models.CharField(max_length=64, unique=True, help_text=_("i386, Itanium, ..."))


    class Meta:
        ordering = ("name", )
        verbose_name_plural = "arches"


    def __unicode__(self):
        return u"%s" % self.name


    def export(self):
        """Export data for xml-rpc."""
        return {
            "id": self.id,
            "name": self.name,
            "pretty_name": self.pretty_name,
        }


    @property
    def worker_count(self):
        if self.id is None:
            return 0
        return Worker.objects.filter(arches__id=self.id).count()


class Channel(models.Model):
    """Model for hub_channel table."""
    name        = models.CharField(max_length=128, help_text=_("Channel name"))


    def __unicode__(self):
        return u"%s" % self.name


    def export(self):
        """Export data for xml-rpc."""
        return {
            "id": self.id,
            "name": self.name,
        }


    @property
    def worker_count(self):
        if self.id is None:
            return 0
        return Worker.objects.filter(channels__id=self.id).count()


class WorkerManager(models.Manager):
    """Custom query manager for Worker model."""

    def enabled(self):
        """Return all enabled workers."""
        return self.filter(enabled=True)


    def ready(self):
        """Return all enabled workers which are ready."""
        return self.filter(enabled=True, ready=True)


class Worker(models.Model):
    """Model for the hub_worker table."""
    worker_key          = models.CharField(max_length=255, unique=True, blank=True, help_text=_("Worker authentication key.<br />Leave blank to generate new key."))
    name                = models.CharField(max_length=128, unique=True, help_text=_("Worker hostname."))
    arches              = models.ManyToManyField(Arch, help_text=_("Supported architectures"))
    channels            = models.ManyToManyField(Channel)
    enabled             = models.BooleanField(default=True, help_text=_("Enabled workers are allowed to process tasks."))
    max_load            = models.PositiveIntegerField(blank=True, default=1, help_text=_("Maximum allowed load (sum of task weights)."))
    max_tasks           = models.PositiveIntegerField(blank=True, default=0, help_text=_("Maximum assigned tasks. (0 = no limit)"))
    min_priority        = models.PositiveIntegerField(default=0, help_text=_("Worker will take only tasks of this or higher priority."))

    # redundant fields to improve performance
    ready               = models.BooleanField(default=True, help_text=_("Is the worker ready to take new tasks?<br />This is a generated field."))
    task_count          = models.PositiveIntegerField(blank=True, default=0, help_text=_("Count of processed tasks.<br />This is a generated field."))
    current_load        = models.PositiveIntegerField(blank=True, default=0, help_text=_("Sum of task weights.<br />This is a generated field."))

    # override default *objects* Manager
    objects = WorkerManager()


    def __unicode__(self):
        return u"%s" % self.name


    def save(self, *args, **kwargs):
        # precompute task count, current load and ready
        tasks = Task.objects.opened().filter(worker=self)
        self.task_count = tasks.count()
        self.current_load = sum(( task.weight for task in tasks if not task.waiting ))
        self.ready = self.enabled and (self.current_load < self.max_load and self.task_count < 3*self.max_load)

        while not self.worker_key:
            # if worker_key is empty, generate a new one
            key = random_string(64)
            if Worker.objects.filter(worker_key=key).count() == 0:
                self.worker_key = key
        super(self.__class__, self).save(*args, **kwargs)


    def export(self):
        """Export data for xml-rpc."""
        return {
            "id": self.id,
            "name": self.name,
            "arches": [ i.export() for i in self.arches.all() ],
            "channels": [ i.export() for i in self.channels.all() ],
            "enabled": self.enabled,
            "max_load": self.max_load,
            "ready": self.ready,
            "task_count": self.task_count,
            "current_load": self.current_load,

            # Add the hub version.
            # This can be used for taskd compatibility checking everytime a worker_info is updated.
            "version": self._get_version(),
        }


    def _get_version(self):
        """Return hub version or None (if settings.VERSION is not set)."""
        return getattr(settings, "VERSION", None)


    def running_tasks(self):
        """Return list of running tasks on this worker."""
        return Task.objects.running().filter(worker=self)


    def assigned_tasks(self):
        """Return list of assigned tasks to this worker."""
        return Task.objects.assigned().filter(worker=self)


    def update_worker(self, enabled, ready, task_count):
        """Update worker attributes. Return worker_info.

        Update only if data provided from a worker differs.
        """
        if (self.enabled, self.ready, self.task_count) != (enabled, ready, task_count):
            self.save()
        return self.export()


class TaskManager(models.Manager):
    """Custom query manager for Task model."""

    def get_query_set(self):
        """Redefine default query set to exclude archived tasks."""
        return models.Manager.get_query_set(self).filter(archive=False)

    def get_and_verify(self, task_id, worker):
        return self.get(id=task_id, worker=worker)

    def running(self):
        """Return list of assigned or opened tasks."""
        return self.filter(state__in=(TASK_STATES["ASSIGNED"], TASK_STATES["OPEN"])).order_by("-exclusive", "id")

    def free(self):
        """Return list of free tasks."""
        return self.filter(state=TASK_STATES["FREE"]).order_by("-exclusive", "id")

    def assigned(self):
        """Return list of assigned tasks."""
        return self.filter(state=TASK_STATES["ASSIGNED"]).order_by("-exclusive", "id")

    def opened(self):
        """Return list of opened tasks."""
        return self.filter(state=TASK_STATES["OPEN"]).order_by("-exclusive", "id")

    def closed(self):
        """Return list of closed tasks."""
        return self.filter(state=TASK_STATES["CLOSED"]).order_by("-exclusive", "id")

    def canceled(self):
        """Return list of canceled tasks."""
        return self.filter(state=TASK_STATES["CANCELED"]).order_by("-exclusive", "id")

    def failed(self):
        """Return list of failed tasks."""
        return self.filter(state=TASK_STATES["FAILED"]).order_by("-exclusive", "id")

    def interrupted(self):
        """Return list of interrupted tasks."""
        return self.filter(state=TASK_STATES["INTERRUPTED"]).order_by("-exclusive", "id")

    def timeout(self):
        """Return list of tasks killed on a timeout."""
        return self.filter(state=TASK_STATES["TIMEOUT"]).order_by("-exclusive", "id")

    def created(self):
        """Return list of created tasks."""
        return self.filter(state=TASK_STATES["CREATED"]).order_by("-exclusive", "id")


class TaskLogs(object):
    """Task log wrapper."""

    def __init__(self, task_obj):
        self.cache = {}
        self.changed = {} # changed logs, will be written on save()
        self.task = task_obj

    def _get_absolute_log_path(self, name):
        task_dir = self.task.task_dir(self.task.id)
        log_path = os.path.abspath(os.path.join(self.task.task_dir(), name))
        if not log_path.startswith(task_dir):
            # make sure log is under task_dir
            raise RuntimeError("Invalid log abspath.")
        return log_path

    def _get_relative_log_path(self, name):
        log_path = os.path.normpath(name)
        if log_path.startswith(".."):
            raise RuntimeError("Invalid log normpath.")
        return log_path

    def _open_log(self, name):
        log_path = self._get_absolute_log_path(name)
        if os.path.isfile(log_path):
            return io.open(log_path, 'rb', LOG_BUFFER_SIZE)
        elif os.path.isfile(log_path + ".gz"):
            out = gzip.open(log_path + ".gz", "rb")

            # GZipFile was not usable with BufferedReader
            # until 2.7
            if sys.version_info[0:2] >= (2, 7):
                out = io.BufferedReader(out, LOG_BUFFER_SIZE)

            return out
        else:
            raise Exception('Cannot find log %s' % name)

    def get_chunk(self, name, offset=0, length=-1):
        """Returns a sequence of bytes from the named log.

        offset -- optional start index of read
        length -- optional length; if omitted, read to end of file

        Will try to avoid returning a chunk ending in the middle of
        a single UTF-8 character, therefore the returned length may
        be less than requested."""

        name = self._get_relative_log_path(name)

        if name in self.cache:
            # already loaded so no point in reading again
            end = None if length < 0 else offset + length
            return self.cache[name][offset:end]

        # not loaded; read just this part
        log_path = self._get_absolute_log_path(name)
        log_file = None
        try:
            log_file = self._open_log(name)
            log_file.seek(offset)
            return _utf8_chunk(log_file.read(length))
        finally:
            if log_file is not None:
                log_file.close()

    def tail(self, name, max_size, max_line_size=8192):
        """Return a byte string containing trailing lines from a log,
        up to a maximum count of bytes.

        name          -- the log name
        max_size      -- returned string will be equal or less than this size
        max_line_size -- max expected size (in bytes) of a line in the file.
                         If lines exceed this size then the return value
                         might not be aligned to the beginning of a line.

        Returns (bytestring, offset) where offset is the total number
        of bytes read from the file (including discarded bytes).
        """
        log_file = None
        try:
            log_file = self._open_log(name)
            return _tail(log_file, max_size, max_line_size)
        finally:
            if log_file is not None:
                log_file.close()

    def __getitem__(self, name):
        """Get full content of named log, as a byte string.

        This method reads (and caches) the entire uncompressed content of the
        log file, thus may cause memory issues if log files are expected to
        be large.  To limit the amount of memory used at once, use the
        get_chunk method instead."""

        name = self._get_relative_log_path(name)
        if name not in self.cache:
            # task.id is still not set. Return empty string.
            if self.task.id is None:
                return ""

            log_path = self._get_absolute_log_path(name)
            if os.path.isfile(log_path):
                self.cache[name] = "\n".join(read_from_file(log_path))
            elif os.path.isfile(log_path + ".gz"):
                fo = gzip.open(log_path + ".gz", "rb")
                self.cache[name] = fo.read()
                fo.close()
            else:
                self.cache[name] = ""
            self.changed[name] = False

        return self.cache[name]

    def __setitem__(self, name, value):
        name = self._get_relative_log_path(name)
        self.cache[name] = value
        self.changed[name] = True

    def save(self):
        for log in self.cache:
            if not self.changed.get(log, True):
                continue
            log_path = self._get_absolute_log_path(log)
            if os.path.basename(log).startswith("traceback"):
                mode = 0600
            else:
                mode = 0644
            save_to_file(log_path, self.cache[log], mode=mode)
            self.changed[log] = False

    @property
    def list(self):
        result = []

        if self.task.id is not None:
            task_dir = self.task.task_dir()
            if not task_dir.endswith("/"):
                task_dir += "/"

            # logs on disk
            for root, dirs, files in os.walk(task_dir):
                for i in files:
                    if i.endswith(".log.gz"):
                        i = i[:-3]
                    result.append(os.path.join(root, i)[len(task_dir):])

        # cached logs
        for log in self.cache:
            if log not in result:
                result.append(log)

        return result

    def _gzip_log(self, name):
        """gzip one log, do *not* throw any error on failure"""

        # compress only log files
        if not name.endswith(".log"):
            return

        import pipes
        path = self._get_absolute_log_path(name)
        if not os.path.isfile(path + ".gz"):
            run("gzip %s" % pipes.quote(path), can_fail=True, stdout=False)

    def gzip_logs(self):
        """gzip all task logs"""
        for i in self.list:
            self._gzip_log(i)


class Task(models.Model):
    """Model for hub_task table."""
    archive             = models.BooleanField(default=False, help_text=_("When a task is archived, it disappears from admin interface and cannot be accessed by taskd.<br />Make sure that archived tasks are finished and you won't need them anymore."))
    owner               = models.ForeignKey(settings.AUTH_USER_MODEL)
    worker              = models.ForeignKey(Worker, null=True, blank=True, help_text=_("A worker which has this task assigned."))
    parent              = models.ForeignKey("self", null=True, blank=True, help_text=_("Parent task."))
    state               = models.PositiveIntegerField(default=TASK_STATES["FREE"], choices=TASK_STATES.get_mapping(), help_text=_("Current task state."))
    label               = models.CharField(max_length=255, blank=True, help_text=_("Label, description or any reason for this task."))
    exclusive           = models.BooleanField(default=False, help_text=_("Exclusive tasks have highest priority. They are used e.g. when shutting down a worker."))

    method              = models.CharField(max_length=255, help_text=_("Method name represents appropriate task handler."))
    args                = kobo.django.fields.JSONField(blank=True, default={}, help_text=_("Method arguments. JSON serialized dictionary."))
    result              = models.TextField(blank=True, help_text=_("Task result. Do not store a lot of data here (use HubProxy.upload_task_log instead)."))
    comment             = models.TextField(null=True, blank=True)

    arch                = models.ForeignKey(Arch)
    channel             = models.ForeignKey(Channel)
    timeout             = models.PositiveIntegerField(null=True, blank=True, help_text=_("Task timeout. Leave blank for no timeout."))

    waiting             = models.BooleanField(default=False, help_text=_("Task is waiting until some subtasks finish."))
    awaited             = models.BooleanField(default=False, help_text=_("Task is awaited by another task."))

    dt_created          = models.DateTimeField(auto_now_add=True)
    dt_started          = models.DateTimeField(null=True, blank=True)
    dt_finished         = models.DateTimeField(null=True, blank=True)

    priority            = models.PositiveIntegerField(default=10, help_text=_("Priority."))
    weight              = models.PositiveIntegerField(default=1, help_text=_("Weight determines how many resources is used when processing the task."))

    resubmitted_by      = models.ForeignKey(settings.AUTH_USER_MODEL, null=True, blank=True, related_name="resubmitted_by1")
    resubmitted_from    = models.ForeignKey("self", null=True, blank=True, related_name="resubmitted_from1")

    subtask_count       = models.PositiveIntegerField(default=0, help_text=_("Subtask count.<br />This is a generated field."))

    # override default *objects* Manager
    objects = TaskManager()

    class Meta:
        ordering = ("-id", )
        permissions = (
            ("can_see_traceback", _("Can see traceback")),
        )

    def __init__(self, *args, **kwargs):
        self.logs = TaskLogs(self)
        traceback = kwargs.pop("traceback", None)
        if traceback:
            self.logs["traceback.log"] = traceback

        stdout = kwargs.pop("stdout", None)
        if stdout:
            self.logs["stdout.log"] = stdout

        super(Task, self).__init__(*args, **kwargs)

    def __unicode__(self):
        if self.parent:
            return u"#%s [method: %s, state: %s, worker: %s, parent: #%s]" % (self.id, self.method, self.get_state_display(), self.worker, self.parent.id)
        return u"#%s [method: %s, state: %s, worker: %s]" % (self.id, self.method, self.get_state_display(), self.worker)

    def save(self, *args, **kwargs):
        # save to db to precalculate subtask counts and obtain an ID (on insert) for stdout and traceback
        self.subtask_count = self.subtasks().count()
        super(self.__class__, self).save()
        self.logs.save()
        if self.parent:
            self.parent.save(*args, **kwargs)

    @classmethod
    def get_task_dir(cls, task_id, create=False):
        '''Task files (logs, etc.) are saved in TASK_DIR in following structure based on task_id:
        TASK_DIR/millions/tens_of_thousands/task_id/*
        '''
        task_id = int(task_id)
        third = task_id
        second = task_id // 10000 * 10000
        first = task_id // 1000000 * 1000000

        task_dir = os.path.abspath(settings.TASK_DIR)
        path = os.path.join(task_dir, str(first), str(second), str(third))
        path = os.path.abspath(path)
        if not path.startswith(task_dir):
            raise Exception('Possible hack, trying to read path "%s"' % path)

        if create and not os.path.isdir(path):
            os.makedirs(path, mode=0755)

        return path

    def task_dir(self, create=False):
        return Task.get_task_dir(self.id, create)

    @classmethod
    def create_task(cls, owner_name, label, method, args=None, comment=None, parent_id=None, worker_name=None, arch_name="noarch", channel_name="default", timeout=None, priority=10, weight=1, exclusive=False, resubmitted_by=None, resubmitted_from=None, state=None):
        """Create a new task."""
        task = cls()
        task.owner = get_user_model().objects.get(username=owner_name)
        task.label = label
        task.method = method
        task.args = args or {}
        task.comment = comment

        if parent_id is not None:
            task.parent = cls.objects.get(id=parent_id)

        if state is not None:
            task.state = state

        if worker_name is not None:
            task.worker = Worker.objects.get(name=worker_name)
            task.state = TASK_STATES["ASSIGNED"]

        task.resubmitted_by = resubmitted_by
        task.resubmitted_from = resubmitted_from

        task.arch = Arch.objects.get(name=arch_name)
        task.channel = Channel.objects.get(name=channel_name)
        task.priority = priority
        task.timeout = timeout
        task.weight = weight
        task.exclusive = exclusive

        # TODO: unsupported in Django 1.0
        #task.validate()
        task.save()
        return task.id

    @classmethod
    def create_shutdown_task(cls, owner_name, worker_name, kill=False):
        """Create a new ShutdownWorker task."""
        kwargs = {
            "owner_name": owner_name,
            "label": "Shutdown a worker.",
            "method": "ShutdownWorker",
            "args": {
                "kill": kill,
            },
            "worker_name": worker_name,
            "weight": 0,
            "exclusive": True,
        }
        return cls.create_task(**kwargs)

    def set_args(self, **kwargs):
        """Serialize args dictionary."""
        print >> sys.stderr, "DeprecationWarning: kobo.hub.models.Task.set_args() is deprecated. Use kobo.hub.models.Task.args instead."
        self.args = kwargs

    def get_args(self):
        """Deserialize args dictionary."""
        print >> sys.stderr, "DeprecationWarning: kobo.hub.models.Task.get_args() is deprecated. Use kobo.hub.models.Task.args instead."
        return self.args.copy()

    def get_args_display(self):
        """Deserialize args dictionary to human readable form"""
        from django.utils.datastructures import SortedDict
        result = SortedDict()
        for key, value in sorted(self.args.items()):
            result[key] = json.dumps(value)
        return result

    def export(self, flat=True):
        """Export data for xml-rpc."""

        result = {
            "id": self.id,
            "owner": self.owner.username,
            "worker": self.worker_id,
            "parent": self.parent_id,
            "state": self.state,
            "label": self.label,

            "method": self.method,
            "args": self.args,
            "result": self.result,

            "exclusive": self.exclusive,
            "arch": self.arch_id,
            "channel": self.channel_id,
            "timeout": self.timeout,
            "waiting": self.waiting,
            "awaited": self.awaited,
            "dt_created": datetime.datetime.strftime(self.dt_created, "%F %R:%S"),
            "dt_started": self.dt_started and datetime.datetime.strftime(self.dt_started, "%Y-%m-%d %H:%M:%S") or None,
            "dt_finished": self.dt_finished and datetime.datetime.strftime(self.dt_finished, "%F %R:%S") or None,
            "priority": self.priority,
            "weight": self.weight,

            "resubmitted_by": getattr(self.resubmitted_by, "username", None),
            "resubmitted_from": getattr(self.resubmitted_from, "id", None),

            # used by task watcher
            "state_label": self.get_state_display(),
            "is_finished": self.is_finished(),
            "is_failed": self.is_failed(),
        }

        if not flat:
            result.update({
                "worker": self.worker and self.worker.export() or None,
                "parent": self.parent and self.parent.export() or None,
                "arch": self.arch.export(),
                "channel": self.channel.export(),
                "subtask_id_list": [ i.id for i in self.subtasks() ],
            })

        return result

    def subtasks(self):
        return Task.objects.filter(parent=self)

    @property
    def time(self):
        """return time spent in the task"""
        if not self.dt_started:
            return None
        elif not self.dt_finished:
            return datetime.datetime.now() - self.dt_started
        else:
            return self.dt_finished - self.dt_started

    def get_time_display(self):
        """display time in human readable form"""
        if self.time is None:
            return ""

        time = "%02d:%02d:%02d" % ((self.time.seconds/60.0/60.0), (self.time.seconds/60.0 % 60), self.time.seconds % 60)
        if self.time.days:
            time = _("%s days, %s") % (self.time.days, time)
        return time
    get_time_display.short_description = "Time"

    def __lock(self, worker_id, new_state=TASK_STATES["ASSIGNED"], initial_states=None):
        """Critical section. Ensures that only one worker takes the task."""

        if type(initial_states) in (list, tuple):
            # filter out invalid state codes
            initial_states = [ i for i, j in TASK_STATES.get_mapping() if i in initial_states ]
            if not initial_states:
                # initial_states is empty
                initial_states = (TASK_STATES["FREE"], )
        else:
            initial_states = (TASK_STATES["FREE"], )

        # it is safe to pass initial_states directly to query,
        # because these values are checked in the code above
        query = """
UPDATE
  hub_task
SET
  state=%%s,
  worker_id=%%s,
  dt_started=%%s,
  dt_finished=%%s,
  waiting=%%s
WHERE
  id=%%s
  and state in (%(initial_states)s)
  and (worker_id is null or worker_id=%%s)
""" % { "initial_states": ",".join(( "'%s'" % i for i in initial_states )), }

        dt_started = self.dt_started
        if new_state == TASK_STATES["OPEN"]:
            dt_started = datetime.datetime.now()

        dt_finished = self.dt_finished
        if new_state in FINISHED_STATES:
            dt_finished = datetime.datetime.now()

        new_worker_id = worker_id
        if new_state == TASK_STATES["FREE"]:
            new_worker_id = None

        waiting = False

        with transaction.atomic():
            cursor = connection.cursor()
            cursor.execute(query, (new_state, new_worker_id, dt_started, dt_finished, waiting, self.id, worker_id))

            if cursor.rowcount == 0:
                if self.state in FINISHED_STATES:
                    logger.debug("Trying to interrupt closed task %s, ignoring.", self.id)
                    return
                else:
                    raise ObjectDoesNotExist()

            if cursor.rowcount > 1:
                raise MultipleObjectsReturned()

        self.dt_started = dt_started
        self.dt_finished = dt_finished
        if new_worker_id is not None:
            self.worker = Worker.objects.get(id=new_worker_id)
        self.state = new_state
        self.waiting = waiting

    def free_task(self):
        """Free the task."""
        try:
            self.__lock(self.worker_id, new_state=TASK_STATES["FREE"], initial_states=(TASK_STATES["FREE"], TASK_STATES["ASSIGNED"], TASK_STATES["CREATED"]))
        except (MultipleObjectsReturned, ObjectDoesNotExist):
            raise Exception("Cannot free task %d, state is %s" % (self.id, self.get_state_display()))

    def assign_task(self, worker_id=None):
        """Assign the task to a worker identified by worker_id."""
        if worker_id is None:
            worker_id = self.worker_id

        try:
            self.__lock(worker_id, new_state=TASK_STATES["ASSIGNED"], initial_states=(TASK_STATES["FREE"], TASK_STATES["CREATED"]))
        except (MultipleObjectsReturned, ObjectDoesNotExist):
            raise Exception("Cannot assign task %d" % (self.id))

    def open_task(self, worker_id=None):
        """Open the task on a worker identified by worker_id."""
        if worker_id is None:
            worker_id = self.worker_id

        try:
            self.__lock(worker_id, new_state=TASK_STATES["OPEN"], initial_states=(TASK_STATES["FREE"], TASK_STATES["ASSIGNED"]))
        except (MultipleObjectsReturned, ObjectDoesNotExist):
            raise Exception("Cannot open task %d, state is %s" % (self.id, self.get_state_display()))

    @transaction.atomic
    def close_task(self, task_result=""):
        """Close the task and save result."""
        if task_result:
            self.result = task_result
            self.save()

        try:
            self.__lock(self.worker_id, new_state=TASK_STATES["CLOSED"], initial_states=(TASK_STATES["OPEN"], ))
        except (MultipleObjectsReturned, ObjectDoesNotExist):
            raise Exception("Cannot close task %d, state is %s" % (self.id, self.get_state_display()))
        self.logs.gzip_logs()

    @transaction.atomic
    def cancel_task(self, user=None, recursive=True):
        """Cancel the task."""
        if user is not None and not user.is_superuser:
            if self.owner.username != user.username:
                raise Exception("You are not task owner or superuser.")

        try:
            self.__lock(self.worker_id, new_state=TASK_STATES["CANCELED"], initial_states=(TASK_STATES["FREE"], TASK_STATES["ASSIGNED"], TASK_STATES["OPEN"], TASK_STATES["CREATED"]))
        except (MultipleObjectsReturned, ObjectDoesNotExist):
            raise Exception("Cannot cancel task %d, state is %s" % (self.id, self.get_state_display()))

        if recursive:
            for task in self.subtasks():
                task.cancel_task(recursive=True)
        self.logs.gzip_logs()

    def cancel_subtasks(self):
        """Cancel all subtasks of the task."""
        result = True
        for task in self.subtasks():
            try:
                result &= task.cancel_task()
            except (MultipleObjectsReturned, ObjectDoesNotExist):
                result = False
        return result

    @transaction.atomic
    def interrupt_task(self, recursive=True):
        """Set the task state to interrupted."""
        try:
            self.__lock(self.worker_id, new_state=TASK_STATES["INTERRUPTED"], initial_states=(TASK_STATES["OPEN"], ))
        except (MultipleObjectsReturned, ObjectDoesNotExist):
            raise Exception("Cannot interrupt task %d, state is %s" % (self.id, self.get_state_display()))

        if recursive:
            for task in self.subtasks():
                task.interrupt_task(recursive=True)
        self.logs.gzip_logs()

    @transaction.atomic
    def timeout_task(self, recursive=True):
        """Set the task state to timeout."""
        try:
            self.__lock(self.worker_id, new_state=TASK_STATES["TIMEOUT"], initial_states=(TASK_STATES["OPEN"], ))
        except (MultipleObjectsReturned, ObjectDoesNotExist):
            raise Exception("Cannot interrupt task %d, state is %s" % (self.id, self.get_state_display()))

        if recursive:
            for task in self.subtasks():
                task.interrupt_task(recursive=True)
        self.logs.gzip_logs()

    @transaction.atomic
    def fail_task(self, task_result=""):
        """Fail this task and save result."""
        if task_result:
            self.result = task_result
            self.save()

        try:
            self.__lock(self.worker_id, new_state=TASK_STATES["FAILED"], initial_states=(TASK_STATES["OPEN"], ))
        except (MultipleObjectsReturned, ObjectDoesNotExist):
            raise Exception("Cannot fail task %i, state is %s" % (self.id, self.get_state_display()))
        self.logs.gzip_logs()

    def is_finished(self):
        """Is the task finished? Task state can be one of: closed, interrupted, canceled, failed."""
        return self.state in FINISHED_STATES

    def is_failed(self):
        """Is the task successfuly finished? Task state must be closed."""
        return self.state in FAILED_STATES

    def resubmit_task(self, user, force=False):
        """Resubmit failed/canceled top-level task."""
        if not user.is_superuser:
            if self.owner.username != user.username:
                raise Exception("You are not task owner or superuser.")

        if self.parent:
            raise Exception("Task is not top-level: %s" % self.id)

        if self.exclusive:
            raise Exception("Cannot resubmit exclusive task: %s" % self.id)

        if not force and self.state not in FAILED_STATES:
            states = [ TASK_STATES.get_value(i) for i in FAILED_STATES ]
            raise Exception("Task '%s' must be in: %s" % (self.id, states))

        kwargs = {
            "owner_name": self.owner.username,
            "label": self.label,
            "method": self.method,
            "args": self.args,
            "comment": self.comment,
            "parent_id": None,
            "worker_name": None,
            "arch_name": self.arch.name,
            "channel_name": self.channel.name,
            "priority": self.priority,
            "weight": self.weight,
            "exclusive": self.exclusive,
            "resubmitted_by": user,
            "resubmitted_from": self,
        }
        return Task.create_task(**kwargs)

    def clone_task(self, user, **kwargs):
        """Clone a task, override field values by kwargs."""
        if not user.is_superuser:
            raise Exception("You are not superuser.")

        if self.parent:
            raise Exception("Task is not top-level: %s" % self.id)

        kwargs.pop("resubmitted_by", None)
        kwargs.pop("resubmitted_from", None)

        new_kwargs = {
            "owner_name": self.owner.username,
            "label": self.label,
            "method": self.method,
            "args": self.args,
            "comment": self.comment,
            "parent_id": None,
            "worker_name": None,
            "arch_name": self.arch.name,
            "channel_name": self.channel.name,
            "priority": self.priority,
            "weight": self.weight,
            "exclusive": self.exclusive,
            "resubmitted_by": user,
            "resubmitted_from": self,
        }

        new_kwargs.update(kwargs)
        return Task.create_task(**new_kwargs)

    def wait(self, child_task_list=None):
        """Set this task as waiting and all subtasks in child_task_list as awaited.

        If child_task_list is None, process all related subtasks.
        """
        tasks = self.subtasks().filter(state__in=(TASK_STATES["FREE"], TASK_STATES["ASSIGNED"], TASK_STATES["OPEN"]))
        if child_task_list is not None:
            tasks = tasks.filter(id__in=child_task_list)

        for task in tasks:
            task.awaited = True
            task.save()

        self.waiting = True
        self.save()

    def check_wait(self, child_task_list=None):
        """Determine if all subtasks have finished."""
        tasks = self.subtasks()
        if child_task_list is not None:
            tasks = tasks.filter(id__in=child_task_list)

        finished = []
        unfinished = []
        for task in tasks:
            if task.is_finished():
                finished.append(task.id)
                task.awaited = False
                task.save()
            else:
                unfinished.append(task.id)

        return [finished, unfinished]


def _task_delete(sender, instance, **kwargs):
    """
    When Task object is deleted, appropriate task_dir is deleted also. This is
    done by catching post_delete signal
    """
    task_dir = Task.get_task_dir(instance.id)    
    try:
        shutil.rmtree(task_dir)
    except OSError:
        pass

post_delete.connect(_task_delete, sender=Task)



#!/usr/bin/python
# -*- coding: utf-8 -*-


import os

import distutils.command.sdist
from distutils.core import setup
from distutils.command.install import INSTALL_SCHEMES


# override default tarball format with bzip2
distutils.command.sdist.sdist.default_format = {"posix": "bztar"}

# force to install data files to site-packages
for scheme in INSTALL_SCHEMES.values():
    scheme["data"] = scheme["purelib"]

# recursively scan for python modules to be included
package_root_dirs = ["kobo"]
packages = set()
package_data = {}
for package_root_dir in package_root_dirs:
    for root, dirs, files in os.walk(package_root_dir):
        # ignore PEP 3147 cache dirs and those whose names start with '.'
        dirs[:] = [i for i in dirs if not i.startswith('.') and i != '__pycache__']
        parts = root.split("/")
        if "__init__.py" in files:
            package = ".".join(parts)
            packages.add(package)
            relative_path = ""
        elif files:
            relative_path = []
            while ".".join(parts) not in packages:
                relative_path.append(parts.pop())
            if not relative_path:
                continue
            relative_path.reverse()
            relative_path = os.path.join(*relative_path)
            package = ".".join(parts)
        else:
            # not a module, no files -> skip
            continue

        package_files = package_data.setdefault(package, [])
        package_files.extend([os.path.join(relative_path, i) for i in files if not i.endswith(".py")])


packages = sorted(packages)
for package in package_data.keys():
    package_data[package] = sorted(package_data[package])


setup(
    name            = "kobo",
    version         = "0.6.0",
    description     = "A pile of python modules used by Red Hat release engineering to build their tools",
    url             = "https://github.com/release-engineering/kobo/",
    author          = "Red Hat, Inc.",
    author_email    = "dmach@redhat.com",
    license         = "LGPLv2.1",

    packages        = packages,
    package_data    = package_data,
    scripts         = ["kobo/admin/kobo-admin"],
)



#!/usr/bin/python
# -*- coding: utf-8 -*-

import json
import unittest

from django.core.exceptions import ValidationError
from .models import DummyDefaultModel, DummyModel, DummyNotHumanModel


class TestBasicJSONField(unittest.TestCase):

    def test_without_default(self):
        """
        This should trigger exception:
            ValidationError: [u"Cannot deserialize JSON data. ''"]
        """
        self.assertRaises(ValidationError, DummyModel)

    def test_default(self):
        """
        Test if default works as it should
        """
        d = DummyDefaultModel()
        d.save()

        self.assertEqual(d.field, {})

    def test_default2(self):
        """
        This should raise 'AssertionError: {} != []'
        """
        d = DummyDefaultModel()
        d.save()

        self.assertNotEqual(d.field, [])

    def test_nothumanreadable(self):
        """
        Test human_readable parameter doesn't change anything on dict level
        """
        d = DummyNotHumanModel()
        d.save()

        self.assertEqual(d.field, {})

    def test_nothumanreadable2(self):
        """
        Test human_readable parameter doesn't change anything with data
        """

        data = {'a': 1}
        d = DummyNotHumanModel()
        d.field = data
        d.save()

        self.assertEqual(data, d.field)


    def test_assignment(self):
        """
        Basic assignment
        """
        d = DummyDefaultModel()
        d.field = []

        self.assertEqual(d.field, [])

    def test_assignment_with_save(self):
        """
        Basic assignment with save
        """
        d = DummyDefaultModel()
        d.field = []
        d.save()

        self.assertEqual(d.field, [])

    def test_complex_assignment_with_save(self):
        d = DummyDefaultModel()
        data = {'asd': [1, 2, 3], 'qwe': {'a': 'b'}}
        d.field = data
        d.save()

        self.assertEqual(d.field, data)


class TestFixturesJSONField(unittest.TestCase):
    """
    DO NOT ADD ANYTHING INTO DATABASE IN THIS TESTCASE

    these tests are meant to test loading fixtures and dumping models
    initial_data.json fixture should be loaded automatically and tested if it's okay
    """
    # will be loaded from <APP>/fixtures/
    #fixtures = ['fixture.json']

    def test_fixture(self):
        """
        Basic assignment
        """
        d = DummyModel.objects.get(pk=1)
        try:
            # python 2.7+
            self.assertIsInstance(d.field, dict)
        except AttributeError:
            # python <2.7
            self.assertTrue(isinstance(d.field, dict))
        self.assertTrue('key' in d.field)
        self.assertEqual(d.field['key'], 'value')

    def test_dump(self):
        """
        Basic assignment
        """
        from django.core import serializers

        JSSerializer = serializers.get_serializer('json')
        js_serializer = JSSerializer()
        js = js_serializer.serialize(DummyModel.objects.all())
        js_des = json.loads(js)
        des_obj = js_des[0]
        self.assertEqual(des_obj['pk'], 1)
        self.assertEqual(json.loads(des_obj['fields']['field']), {'key': 'value', })




#!/usr/bin/python
# -*- coding: utf-8 -*-


import unittest2 as unittest
import run_tests # set sys.path

import os
import shutil
import tempfile
from cStringIO import StringIO

from kobo.shortcuts import *


class TestShortcuts(unittest.TestCase):
    def test_force_list(self):
        self.assertEqual(force_list("a"), ["a"])
        self.assertEqual(force_list(["a"]), ["a"])
        self.assertEqual(force_list(["a", "b"]), ["a", "b"])
        self.assertItemsEqual(force_list(set(["a", "b"])), ["a", "b"])

    def test_force_tuple(self):
        self.assertItemsEqual(force_tuple("a"), ("a",))
        self.assertItemsEqual(force_tuple(("a",)), ("a",))
        self.assertItemsEqual(force_tuple(("a", "b")), ("a", "b"))
        self.assertItemsEqual(force_tuple(set(["a", "b"])), ("a", "b"))

    def test_allof(self):
        self.assertEqual(allof(), True)
        self.assertEqual(allof(1), True)
        self.assertEqual(allof(True), True)
        self.assertEqual(allof(True, 1, "a"), True)
        self.assertEqual(allof(0), False)
        self.assertEqual(allof(""), False)
        self.assertEqual(allof(None), False)

    def test_anyof(self):
        self.assertEqual(anyof(), False)
        self.assertEqual(anyof(1), True)
        self.assertEqual(anyof(True), True)
        self.assertEqual(anyof(True, 0, "a"), True)
        self.assertEqual(anyof(0), False)
        self.assertEqual(anyof(""), False)
        self.assertEqual(anyof(None), False)

    def test_noneof(self):
        self.assertEqual(noneof(), True)
        self.assertEqual(noneof(False), True)
        self.assertEqual(noneof(True), False)
        self.assertEqual(noneof(False, "", 0), True)
        self.assertEqual(noneof(True, "a", 1), False)
        self.assertEqual(noneof(False, "a", 1), False)
        self.assertEqual(noneof(0, True, False, "a"), False)

    def test_oneof(self):
        self.assertEqual(oneof(), False)
        self.assertEqual(oneof(True), True)
        self.assertEqual(oneof(False), False)
        self.assertEqual(oneof(0, False, "a"), True)
        self.assertEqual(oneof(0, True, False, "a"), False)
        self.assertEqual(oneof(1, True, "a"), False)
        self.assertEqual(oneof(0, False, ""), False)

    def test_is_empty(self):
        self.assertEqual(is_empty(None), True)
        self.assertEqual(is_empty([]), True)
        self.assertEqual(is_empty([1]), False)
        self.assertEqual(is_empty(()), True)
        self.assertEqual(is_empty((1,)), False)
        self.assertEqual(is_empty({}), True)
        self.assertEqual(is_empty(1), False)

    def test_iter_chunks(self):
        self.assertEqual(list(iter_chunks([], 100)), [])
        self.assertEqual(list(iter_chunks(range(5), 1)), [[0], [1], [2], [3], [4]])
        self.assertEqual(list(iter_chunks(range(5), 2)), [[0, 1], [2, 3], [4]])
        self.assertEqual(list(iter_chunks(range(5), 5)), [[0, 1, 2, 3, 4]])
        self.assertEqual(list(iter_chunks(range(6), 2)), [[0, 1], [2, 3], [4, 5]])

        self.assertEqual(list(iter_chunks(xrange(5), 2)), [[0, 1], [2, 3], [4]])
        self.assertEqual(list(iter_chunks(xrange(6), 2)), [[0, 1], [2, 3], [4, 5]])
        self.assertEqual(list(iter_chunks(xrange(1, 6), 2)), [[1, 2], [3, 4], [5]])
        self.assertEqual(list(iter_chunks(xrange(1, 7), 2)), [[1, 2], [3, 4], [5, 6]])

        def gen(num):
            for i in xrange(num):
                yield i+1
        self.assertEqual(list(iter_chunks(gen(5), 2)), [[1, 2], [3, 4], [5]])

        self.assertEqual(list(iter_chunks("01234", 2)), ["01", "23", "4"])
        self.assertEqual(list(iter_chunks("012345", 2)), ["01", "23", "45"])

        file_obj = open(os.path.dirname(os.path.abspath(__file__)) + "/chunks_file", "r")
        self.assertEqual(list(iter_chunks(file_obj, 11)), (10 * ["1234567890\n"]) + ["\n"])

        string_io = StringIO((10 * "1234567890\n") + "\n")
        self.assertEqual(list(iter_chunks(string_io, 11)), (10 * ["1234567890\n"]) + ["\n"])


class TestUtils(unittest.TestCase):
    def setUp(self):
        self.tmp_dir = tempfile.mkdtemp()
        self.tmp_file = os.path.join(self.tmp_dir, "tmp_file")
        save_to_file(self.tmp_file, "test")

    def tearDown(self):
        shutil.rmtree(self.tmp_dir)

    def test_save_to_file(self):
        save_to_file(self.tmp_file, "foo")
        self.assertEqual("\n".join(read_from_file(self.tmp_file)), "foo")

        save_to_file(self.tmp_file, "\nbar", append=True, mode=600)
        self.assertEqual("\n".join(read_from_file(self.tmp_file)), "foo\nbar")

        # append doesn't modify existing perms
        self.assertEqual(os.stat(self.tmp_file).st_mode & 0777, 0644)

        os.unlink(self.tmp_file)
        save_to_file(self.tmp_file, "foo", append=True, mode=0600)
        self.assertEqual(os.stat(self.tmp_file).st_mode & 0777, 0600)

    def test_run(self):
        ret, out = run("echo hello")
        self.assertEqual(ret, 0)
        self.assertEqual(out, "hello\n")

        ret, out = run(["echo", "'hello'"])
        self.assertEqual(ret, 0)
        self.assertEqual(out, "'hello'\n")

        ret, out = run(["echo", "\" ' "])
        self.assertEqual(ret, 0)
        self.assertEqual(out, "\" ' \n")

        # test a longer output that needs to be read in several chunks
        ret, out = run("echo -n '%s'; sleep 0.2; echo -n '%s'" % (10000 * "x", 10 * "a"), logfile=self.tmp_file, can_fail=True)
        self.assertEqual(ret, 0)
        self.assertEqual(out, 10000 * "x" + 10 * "a")
        # check if log file is written properly; it is supposed to append data to existing content
        self.assertEqual("\n".join(read_from_file(self.tmp_file)), "test" + 10000 * "x" + 10 * "a")

        ret, out = run("exit 1", can_fail=True)
        self.assertEqual(ret, 1)

        self.assertRaises(RuntimeError, run, "exit 1")

        # stdin test
        ret, out = run("xargs -0 echo -n", stdin_data="\0".join([str(i) for i in xrange(10000)]))
        self.assertEqual(out, " ".join([str(i) for i in xrange(10000)]))

        # return None
        ret, out = run("xargs echo", stdin_data="\n".join([str(i) for i in xrange(1000000)]), return_stdout=False)
        self.assertEqual(out, None)

        # log file with absolute path
        log_file = os.path.join(self.tmp_dir, "a.log")
        ret, out = run("echo XXX", logfile=log_file)
        self.assertEqual(open(log_file, "r").read(), "XXX\n")

        # log file with relative path
        log_file = "b.log"
        cwd = os.getcwd()
        os.chdir(self.tmp_dir)
        ret, out = run("echo XXX", logfile=log_file)
        self.assertEqual(open(log_file, "r").read(), "XXX\n")
        os.chdir(cwd)

        # bashism - output redirection to subshells
        # fails in default shell (/bin/sh)
        self.assertRaises(RuntimeError, run, "echo foo | tee >(md5sum -b) >/dev/null")
        # passes in bash
        run("echo foo | tee >(md5sum -b) >/dev/null", executable="/bin/bash")

    def test_run_move_dir_with_logfile(self):
        # Write log into a directory that gets renamed by the called command.
        # The file should be opened just once at the start.
        tmpdir = tempfile.mkdtemp(prefix='kobo-bug-')
        logfile = os.path.join(tmpdir, 'file.log')
        destdir = tmpdir + '.moved'
        run(['mv', '-v', tmpdir, destdir], logfile=logfile, show_cmd=True)
        self.assertTrue(os.path.isfile(os.path.join(destdir, 'file.log')))

    def test_read_checksum_file(self):
        data = r"""01186fcf04b4b447f393e552964c08c7b419c1ad7a25c342a0b631b1967d3a27 *test-data/a b
a63d8014dba891345b30174df2b2a57efbb65b4f9f09b98f245d1b3192277ece *test-data/ab
\911169ddaaf146aff539f58c26c489af3b892dff0fe283c1c264c65ae5aa59a2 *test-data/a\nb
ef743c494c8ed766272eef7992607a843799149252822266adc302547587253d *test-data/a"b
\eaba35b63f3a21c43bc4d579fa4ae0cd388ec8633c08e0a54859d07d33a0c487 *test-data/a\\b"""
        open(self.tmp_file, "w").write(data)

        checksums = read_checksum_file(self.tmp_file)

        checksum, path = checksums[0]
        self.assertEqual(checksum, "01186fcf04b4b447f393e552964c08c7b419c1ad7a25c342a0b631b1967d3a27")
        self.assertEqual(path, "test-data/a b")

        checksum, path = checksums[1]
        self.assertEqual(checksum, "a63d8014dba891345b30174df2b2a57efbb65b4f9f09b98f245d1b3192277ece")
        self.assertEqual(path, "test-data/ab")

        checksum, path = checksums[2]
        self.assertEqual(checksum, "911169ddaaf146aff539f58c26c489af3b892dff0fe283c1c264c65ae5aa59a2")
        self.assertEqual(path, "test-data/a\nb")

        checksum, path = checksums[3]
        self.assertEqual(checksum, "ef743c494c8ed766272eef7992607a843799149252822266adc302547587253d")
        self.assertEqual(path, "test-data/a\"b")

        checksum, path = checksums[4]
        self.assertEqual(checksum, "eaba35b63f3a21c43bc4d579fa4ae0cd388ec8633c08e0a54859d07d33a0c487")
        self.assertEqual(path, "test-data/a\\b")

    def test_compute_file_checksums(self):
        self.assertEqual(compute_file_checksums(self.tmp_file, "md5"), dict(md5="098f6bcd4621d373cade4e832627b4f6"))
        self.assertEqual(compute_file_checksums(self.tmp_file, ["md5", "sha256"]), dict(md5="098f6bcd4621d373cade4e832627b4f6", sha256="9f86d081884c7d659a2feaa0c55ad015a3bf4f1b2b0b822cd15d6c15b0f00a08"))
        self.assertEqual(compute_file_checksums(self.tmp_file, ["md5", "md5"]), dict(md5="098f6bcd4621d373cade4e832627b4f6"))
        self.assertRaises(ValueError, compute_file_checksums, self.tmp_file, "unsupported_checksum")

    def test_makedirs(self):
        path = os.path.join(self.tmp_dir, "dir")
        makedirs(path)
        makedirs(path)

        path = os.path.join(self.tmp_dir, "file")
        open(path, "w").close()
        self.assertRaises(OSError, makedirs, path)


class TestPaths(unittest.TestCase):
    def test_split_path(self):
        self.assertEqual(split_path(""), ["."])
        self.assertEqual(split_path("../"), [".."])
        self.assertEqual(split_path("/"), ["/"])
        self.assertEqual(split_path("//"), ["/"])
        self.assertEqual(split_path("///"), ["/"])
        self.assertEqual(split_path("/foo"), ["/", "foo"])
        self.assertEqual(split_path("/foo/"), ["/", "foo"])
        self.assertEqual(split_path("/foo//"), ["/", "foo"])
        self.assertEqual(split_path("/foo/bar"), ["/", "foo", "bar"])
        self.assertEqual(split_path("/foo//bar"), ["/", "foo", "bar"])

    def test_relative_path(self):
        self.assertEqual(relative_path("/foo", "/"), "foo")
        self.assertEqual(relative_path("/foo/", "/"), "foo/")
        self.assertEqual(relative_path("/foo", "/bar/"), "../foo")
        self.assertEqual(relative_path("/foo/", "/bar/"), "../foo/")
        self.assertEqual(relative_path("/var/www/template/index.html", "/var/www/html/index.html"), "../template/index.html")
        self.assertEqual(relative_path("/var/www/template/index.txt", "/var/www/html/index.html"), "../template/index.txt")
        self.assertEqual(relative_path("/var/www/template/index.txt", "/var/www/html/index.html"), "../template/index.txt")
        self.assertRaises(RuntimeError, relative_path, "/var/www/template/", "/var/www/html/index.html")


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

"""
Run tests in current directory.
"""

import os
import sys


# prepare environment for tests
PROJECT_DIR = os.path.abspath(os.path.join(os.path.dirname(__file__), ".."))
os.environ["PYTHONPATH"] = PROJECT_DIR
os.environ['DJANGO_SETTINGS_MODULE'] = 'settings'
sys.path.insert(0, PROJECT_DIR)


from kobo.shortcuts import run


def run_test(test_with_args):
    print "Executing tests in %-40s" % test_with_args[0]
    retcode, output = run(['python'] + test_with_args, can_fail=True)
    if retcode == 0:
        # If tests were skipped, give caller a hint that this happened.
        # Summary line example:
        # "OK (skipped=9)"
        if 'skipped=' in output.splitlines()[-1]:
            print "[  SKIP  ]"
        else:
            print "[   OK   ]"
        return True
    else:
        print "[ FAILED ]"
        print output
        return False


def run_discovered_tests():
    ok = True
    for test in sorted(os.listdir('.')):
        # run all tests that match the 'test_*.py" pattern
        if not test.startswith("test_"):
            continue
        if not test.endswith(".py"):
            continue

        if not run_test([test]):
            ok = False

    return ok


if __name__ == '__main__':
    os.chdir(os.path.dirname(__file__))
    if sys.argv[1:]:
        # arguments passed; assume it's a test, just run it
        ok = run_test(sys.argv[1:])
    else:
        # run all the tests in the directory
        ok = run_discovered_tests()

    sys.exit(0 if ok else 1)



#!/usr/bin/python
# -*- coding: utf-8 -*-


"""
Notification module.
"""


import smtplib
import sys
import optparse

import kobo.shortcuts


class EmailNotification(object):
    """Send notification e-mails."""

    def __init__(self, smtp_host):
        # connect to SMTP server
        self.smtp_host = smtp_host
        self.server = smtplib.SMTP(smtp_host)

    def __del__(self):
        # disconnect from SMTP server
        self.server.quit()

    def send(self, from_addr, recipients, subject, body, reply_to=None, xheaders=None):
        """send a notification"""
        recipients = kobo.shortcuts.force_list(recipients)
        xheaders = xheaders or {}

        for to_addr in recipients:
            headers = []
            headers.append("From: %s" % from_addr)
            headers.append("Subject: %s" % subject)
            headers.append("To: %s" % to_addr)
            if reply_to:
                headers.append("Reply-To: %s" % reply_to)

            for key, value in xheaders.iteritems():
                if not key.startswith("X-"):
                    raise KeyError("X-Header has to start with 'X-': %s" % key)
                headers.append("%s: %s" % (key, value))

            headers.append("") # blank line after headers
            headers.append(body)

            message = "\r\n".join(headers)
            self.server.sendmail(from_addr, to_addr, message)


def main(argv):
    """Main function for command line usage"""
    parser = optparse.OptionParser("usage: %prog <options> <to_addr> [to_addr]...")
    parser.add_option(
        "--server",
        help="specify SMTP server address"
    )
    parser.add_option(
        "-f",
        "--from",
        dest="from_addr",
        help="set the From address"
    )
    parser.add_option(
        "-s",
        "--subject",
        help="set email Subject"
    )
    parser.add_option(
        "-r",
        "--reply-to",
        help="set the Reply-To address"
    )
    parser.add_option(
        "-x",
        "--xheader",
        nargs=2,
        dest="xheaders",
        action="append",
        help="set X-Headers; takes two arguments (-x X-Spam eggs)"
    )

    (opts, args) = parser.parse_args(argv)

    server = opts.server
    from_addr = opts.from_addr
    subject = opts.subject
    reply_to = opts.reply_to
    xheaders = opts.xheaders and dict(opts.xheaders) or {}
    recipients = args

    if not server:
        parser.error("SMTP server address required")

    if not from_addr or "@" not in from_addr:
        parser.error("invalid From address: %s" % from_addr)

    if not subject:
        parser.error("empty Subject")

    if len(recipients) == 0:
        parser.error("at least one recipient required")

    for to_addr in recipients:
        if "@" not in to_addr:
            parser.error("invalid To address: %s" % to_addr)

    for key in xheaders:
        if not key.startswith("X-"):
            parser.error("X-Header has to start with 'X-': %s" % key)

    notify = EmailNotification(server)
    body = sys.stdin.read()
    notify.send(from_addr, recipients, subject, body, reply_to=reply_to, xheaders=xheaders)


if __name__ == "__main__":
    main(sys.argv[1:])



# -*- coding: utf-8 -*-


from kobo.worker import TaskBase
from kobo.exceptions import ShutdownException


class ShutdownWorker(TaskBase):
    enabled = True

    arches = ["noarch"]
    channels = ["default"]
    exclusive = True
    foreground = True
    priority = 19
    weight = 0.0

    def run(self):
        kill = self.args.get("kill", False)

        if kill:
            # raise exception and terminate immediately
            raise ShutdownException()

        # lock the task manager and let it terminate all tasks
        self.task_manager.locked = True
